from arclet.alconna import Alconna, Arparma
from arclet.letoderea import Contexts, Result, make_event, provide

from ..message import MessageChain
from ..session import Session
from .base import Reply


@make_event(name="entari.event/command/execute")
class CommandExecute:
    message: str | MessageChain

    async def gather(self, context: Contexts):
        if isinstance(self.message, str):
            context["$message"] = MessageChain(self.message)
        else:
            context["$message"] = self.message

    providers = [provide(MessageChain, call="$message")]

    def check_result(self, value) -> Result[str | MessageChain] | None:
        if isinstance(value, str | MessageChain):
            return Result(value)


@make_event(name="entari.event/command/output")
class CommandOutput:
    """依据输出信息的类型，将字符串转换为消息对象以便发送。"""

    session: Session
    command: Alconna
    type: str
    content: str

    async def gather(self, context: Contexts):
        context["$session"] = self.session
        context["$out_command"] = self.command
        context["$out_content"] = self.content

    providers = [
        provide(Alconna, call="$out_command"),
        provide(str, call="$out_content"),
    ]

    def check_result(self, value) -> Result[str | bool | MessageChain] | None:
        if isinstance(value, str | bool | MessageChain):
            return Result(value)


@make_event(name="entari.event/command/before_parse")
class CommandReceive:
    """尝试解析命令时调用"""

    session: Session
    command: Alconna
    content: MessageChain
    reply: Reply | None = None

    async def gather(self, context: Contexts):
        context["$session"] = self.session
        context["$cmd_content"] = self.content
        context["$cmd_command"] = self.command
        if self.reply:
            context["$cmd_reply"] = self.reply

    providers = [
        provide(MessageChain, call="$cmd_content"),
        provide(Alconna, call="$cmd_command"),
        provide(Reply, call="$cmd_reply"),
    ]

    def check_result(self, value) -> Result[MessageChain] | None:
        if isinstance(value, MessageChain):
            return Result(value)


@make_event(name="entari.event/command/after_parse")
class CommandParse:
    """解析完命令后调用"""

    session: Session
    command: Alconna
    result: Arparma

    async def gather(self, context: Contexts):
        context["$session"] = self.session
        context["$parsed_command"] = self.command
        context["$parsed_result"] = self.result

    providers = [
        provide(Alconna, call="$parsed_command"),
        provide(Arparma, call="$parsed_result"),
    ]

    def check_result(self, value) -> Result[Arparma | bool] | None:
        if isinstance(value, Arparma | bool):
            return Result(value)



from __future__ import annotations

import fnmatch
import inspect
import logging
from pathlib import Path
import re
import sys
import traceback
from types import TracebackType
from typing import TYPE_CHECKING, cast

from arclet.letoderea.exceptions import ExceptionHandler, Trace
from loguru import logger
from satori.parser import snake_case

if TYPE_CHECKING:
    from loguru import Logger, Record


TRACE_NO = logger.level("TRACE").no
DEBUG_NO = logger.level("DEBUG").no


def escape_tag(s: str) -> str:
    """用于记录带颜色日志时转义 `<tag>` 类型特殊标签

    参考: [loguru color 标签](https://loguru.readthedocs.io/en/stable/api/logger.html#color)

    参数:
        s: 需要转义的字符串
    """
    return re.sub(r"</?((?:[fb]g\s)?[^<>\s]*)>", r"\\\g<0>", s)


class LoggerManager:
    def __init__(self):
        self.loggers: dict[str, Logger] = {}
        self.fork("[core]")
        self.fork("[plugin]")
        self.fork("[message]")
        self.fork("[error]")
        self.log_level = "INFO"
        self.levelno = logger.level("INFO").no
        self.ignores = set()
        self.short_level = False

    def fork(self, child_name: str):
        patched = logger.patch(lambda r: r.update(name=child_name))
        patched = patched.bind(name=child_name)
        self.loggers[child_name] = patched
        return patched

    @property
    def core(self):
        return self.loggers["[core]"].opt(colors=True)

    @property
    def plugin(self):
        return self.loggers["[plugin]"].opt(colors=True)

    @property
    def message(self):
        return self.loggers["[message]"]

    @property
    def error(self):
        return self.loggers["[error]"]

    def wrapper(self, name: str, color: str = "blue"):
        name = escape_tag(name)
        if name.startswith("[") and name.endswith("]"):
            name1 = f"[{snake_case(name[1:-1])}]"
        else:
            name1 = snake_case(name)
        patched = logger.patch(lambda r: r.update(name=name1, extra=r["extra"] | {"entari_plugin_color": color}))
        patched = patched.bind(name=f"plugins.{name1}")
        self.loggers[name] = patched
        return patched

    def set_level(self, level: str | int):
        if isinstance(level, str):
            level = level.upper()
        logging.basicConfig(
            handlers=[LoguruHandler()],
            level="NOTSET" if level == "TRACE" else level,
            format="%(asctime)s | %(name)s[%(levelname)s]: %(message)s",
            force=True,
        )
        self.log_level = level
        self.levelno = logger.level(level).no if isinstance(level, str) else level


class LoguruHandler(logging.Handler):  # pragma: no cover
    """logging 与 loguru 之间的桥梁，将 logging 的日志转发到 loguru。"""

    def emit(self, record: logging.LogRecord):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


logging.basicConfig(
    handlers=[LoguruHandler()],
    level="INFO",
    format="%(asctime)s | %(name)s[%(levelname)s]: %(message)s",
    force=True,
)

log = LoggerManager()


def default_filter(record):
    if any(fnmatch.fnmatch(record["name"], pattern) for pattern in log.ignores):
        return False
    if record["name"].startswith("launart"):
        if log.levelno <= TRACE_NO:
            return record["level"].no >= logger.level("TRACE").no
        elif log.levelno <= DEBUG_NO:
            return record["level"].no >= logger.level("SUCCESS").no
        else:
            return record["level"].no > logger.level("SUCCESS").no
    return record["level"].no >= log.levelno


def _custom_format(record: Record):
    color = record["extra"].get("entari_plugin_color", "m")
    if record["name"].startswith("[") and record["name"].endswith("]"):  # type: ignore
        name = f"<{color}>{{name}}</{color}>"
    else:
        name = f"<{color}><u>{{name}}</u></{color}>"
    if log.levelno <= TRACE_NO:
        time = "<lk>{time:YYYY-MM-DD HH:mm:ss.SSS}</lk>"
    elif log.levelno <= DEBUG_NO:
        time = "<lk>{time:YYYY-MM-DD HH:mm:ss}</lk>"
    else:
        time = "<lk>{time:MM-DD HH:mm:ss}</lk>"
    level = "<lvl>{level:<7}</lvl> │"
    if log.short_level:
        level = "[<lvl>{level.name[0]}</lvl>]"
    res = f"{time} {level} {name} <lvl>{{message}}</lvl>\n"
    if record["exception"]:
        res += "{exception}\n"
    return res


logger.remove()
logger_id = logger.add(
    sys.stdout,
    level=0,
    diagnose=True,
    backtrace=False,
    colorize=True,
    filter=default_filter,
    format=_custom_format,
)
"""默认日志处理器 id"""


_HIDDEN_UPSTREAMS = {
    "satori": "satori",
    "launart": "launart",
    "uvicorn": "uvicorn",
    "starlette": "starlette",
    "graia.amnesia": "amnesia",
}


def add_hidden_upstream(prefix: str, name: str):
    """添加隐藏上游日志源

    Args:
        prefix (str): 日志源前缀
        name (str): 日志源名称
    """
    _HIDDEN_UPSTREAMS[prefix] = name


def _hidden_upsteam(record: Record):
    for prefix, name in _HIDDEN_UPSTREAMS.items():
        if record["name"].startswith(prefix):  # type: ignore
            record["name"] = name


logger.configure(patcher=_hidden_upsteam)


def loguru_exc_callback(cls: type[BaseException], val: BaseException, tb: TracebackType | None, *_, **__):
    """loguru 异常回调

    Args:
        cls (Type[Exception]): 异常类
        val (Exception): 异常的实际值
        tb (TracebackType): 回溯消息
    """
    log.error.opt(exception=(cls, val, tb)).error("Exception:")


def loguru_print_trace(te: Trace):
    summary = te.stack[0]

    class FakeFrame:
        f_code = type(
            "code", (), {"co_filename": summary.filename, "co_name": summary.name, "co_firstlineno": summary.lineno}
        )
        f_lineno = summary.lineno
        f_globals = {}
        f_locals = {}
        f_builtins = {}

    class FakeTraceback:
        tb_frame = FakeFrame()
        tb_lineno = summary.lineno
        tb_next = te.exc_traceback

    log.error.opt(exception=(te.exc_type, te.exc_value, cast(TracebackType, FakeTraceback()))).error("Exception:")


traceback.print_exception = loguru_exc_callback
ExceptionHandler.print_trace = loguru_print_trace
sys.excepthook = loguru_exc_callback


def apply_log_save(
    log_dir: str | Path,
    rotation: str = "00:00",
    compression: str | None = "zip",
    colorize: bool = False,
):
    log_id = logger.add(
        f"{log_dir}/latest.log",
        level=0,
        enqueue=False,
        rotation=rotation,
        compression=compression,
        colorize=colorize,
        diagnose=True,
        backtrace=True,
        filter=default_filter,
        format=_custom_format,
    )
    return lambda: logger.remove(log_id)


try:
    from loguru._better_exceptions import ExceptionFormatter
    from rich.console import Console
    from rich.traceback import Frame, Traceback

    csl = Console(stderr=False, force_terminal=True)

    def patch_format_exception(self, type_, value, tb, *, from_decorator=False):
        rich_tb = Traceback.from_exception(
            type_,
            value,
            tb,
            show_locals=True,
            width=csl.width,
            code_width=120 if csl.width > 120 else 88,
        )
        segments = [*csl.render(rich_tb)]
        yield csl._render_buffer(segments)

    def patch_print_trace(te: Trace):
        trace = Traceback.extract(te.exc_type, te.exc_value, te.exc_traceback, show_locals=True)
        summary = te.stack[0]
        trace.stacks[0].frames.insert(0, Frame(summary.filename, summary.lineno, summary.name, summary.line, locals={}))  # type: ignore
        rich_tb = Traceback(trace, width=csl.width, code_width=120 if csl.width > 120 else 88)
        segments = [*csl.render(rich_tb)]
        log.error.opt().error(f"Exception:\n{''.join(csl._render_buffer(segments))}")

    def enable_rich_except():
        ExceptionFormatter.format_exception = patch_format_exception
        ExceptionHandler.print_trace = patch_print_trace

except ImportError:

    def enable_rich_except():
        log.error.warning("无法启用 rich except，未安装 rich 库，请通过 `pip install rich` 安装。")


__all__ = ["log", "logger_id", "apply_log_save", "escape_tag", "enable_rich_except", "add_hidden_upstream"]



from arclet.letoderea import make_event, provide
from satori.client import Account
from satori.model import LoginStatus


@make_event(name="entari.event/startup")
class Startup:
    pass


@make_event(name="entari.event/ready")
class Ready:
    pass


@make_event(name="entari.event/cleanup")
class Cleanup:
    pass


@make_event(name="entari.event/account_update")
class AccountUpdate:
    account: Account
    status: LoginStatus

    providers = [provide(Account, call="account")]



import asyncio
from dataclasses import asdict
from pathlib import Path

from arclet.letoderea import post, publish
from launart import Launart, Service, any_completed
from launart.status import Phase
from loguru import logger as loguru_logger

try:
    from watchfiles import PythonFilter, awatch
except ModuleNotFoundError:
    raise ImportError("Please install `watchfiles` first. Install with `pip install arclet-entari[reload]`")

from arclet.entari import add_service, load_plugin, metadata, plugin_config
from arclet.entari.config import BasicConfModel, EntariConfig, model_field
from arclet.entari.event.config import ConfigReload
from arclet.entari.logger import log
from arclet.entari.plugin import find_plugin, find_plugin_by_file, unload_plugin_async

# declare_static()
loguru_logger.disable("watchfiles.main")


class Config(BasicConfModel):
    watch_dirs: list[str | Path] = model_field(
        default_factory=lambda: ["."],
        description="需要监视的目录列表，支持相对路径和绝对路径",
    )
    watch_config: bool = model_field(default=False, description="是否监视配置文件的变化，默认为 False")
    debounce: int = model_field(
        default=1600, description="防抖时间，单位为毫秒，防止频繁变动导致的重复加载，默认为 1600 毫秒"
    )
    step: int = model_field(default=50, description="轮询间隔，单位为毫秒，默认为 50 毫秒")


metadata(
    "Auto Reload",
    author=[{"name": "RF-Tar-Railt", "email": "rf_tar_railt@qq.com"}],
    description="自动监视指定目录下的 Python 或配置文件变化并重新加载对应插件",
    config=Config,
    readme="""
# Auto Reload (Hot Module Replacement)

该插件提供自动监视指定目录下的 Python 或配置文件变化并重新加载对应插件的功能。

## 配置

- `watch_dirs`: 需要监视的目录列表，支持相对路径和绝对路径，默认为当前目录 `["."]`
- `watch_config`: 是否监视配置文件的变化，默认为 `False`
- `debounce`: 防抖时间，单位为毫秒，防止频繁变动导致的重复加载，默认为 `1600` 毫秒
- `step`: 轮询间隔，单位为毫秒，默认为 `50` 毫秒

## 说明

1. 若插件被标记为静态 (通过 `declare_static()`, 或作为 prelude 插件)，其变动将被忽略。
2. 配置文件变动时：
    - auto_reload 会发布 `ConfigReload` 事件，其他插件可监听该事件以处理配置变动。
    - 若目标插件确认配置变动已被自身处理 (通过返回 `True`)，则不会重新加载该插件。
    - 若目标插件未处理配置变动，且非静态插件，则会尝试重新加载该插件。
    - 若插件配置在变动中被移除，则会卸载该插件。
""",
)


logger = log.wrapper("[AutoReload]").opt(colors=True)


class Watcher(Service):
    id = "entari.plugin.auto_reload/watcher"

    @property
    def required(self) -> set[str]:
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "blocking", "cleanup"}

    def __init__(self, config: Config):
        self.config = config
        self.fail: dict[str, tuple[str, dict]] = {}
        super().__init__()

    async def watch(self):
        async for event in awatch(
            *self.config.watch_dirs, debounce=self.config.debounce, step=self.config.step, watch_filter=PythonFilter()
        ):
            for change in event:
                if plugin := find_plugin_by_file(change[1]):
                    if plugin.is_static:
                        logger.info(f"Plugin <y>{plugin.id!r}</y> is static, ignored.")
                        continue
                    logger.info(f"Detected change in <blue>{plugin.id!r}</blue>, reloading...")
                    pid = plugin.id
                    _conf = plugin.config.copy()
                    del plugin
                    await unload_plugin_async(pid)
                    if plugin := load_plugin(pid, _conf):
                        logger.info(f"Reloaded <blue>{plugin.id!r}</blue>")
                        del plugin
                    else:
                        logger.error(f"Failed to reload <blue>{pid!r}</blue>")
                        self.fail[change[1]] = (pid, _conf)
                elif change[1] in self.fail:
                    logger.info(f"Detected change in {change[1]!r} which failed to reload, retrying...")
                    if plugin := load_plugin(*self.fail[change[1]]):
                        logger.info(f"Reloaded <blue>{plugin.id!r}</blue>")
                        del plugin
                        del self.fail[change[1]]
                    else:
                        logger.error(f"Failed to reload <blue>{self.fail[change[1]][0]!r}</blue>")

    async def watch_config(self):
        file = EntariConfig.instance.path.resolve()
        extra = [file.parent.joinpath(path) for path in EntariConfig.instance.plugin_extra_files]
        async for event in awatch(
            file.absolute(),
            *(dir_.absolute() for dir_ in extra),
            debounce=self.config.debounce,
            step=self.config.step,
            recursive=False,
        ):
            for change in event:
                if not self.config.watch_config:
                    continue
                if not (
                    (change[0].name == "modified" and Path(change[1]).resolve() == file)
                    or Path(change[1]).resolve() in extra
                    or Path(change[1]).resolve().parent in extra
                ):
                    continue
                old_basic = asdict(EntariConfig.instance.basic)  # noqa
                old_plugin = EntariConfig.instance.plugin.copy()
                if not EntariConfig.instance.reload():
                    continue
                logger.info(f"Detected change in {change[1]!r}, reloading config...")
                new_basic = asdict(EntariConfig.instance.basic)  # noqa
                for key in old_basic:
                    if key in new_basic and old_basic[key] != new_basic[key]:
                        logger.debug(
                            f"Basic config <y>{key!r}</y> changed from <r>{old_basic[key]!r}</r> "
                            f"to <g>{new_basic[key]!r}</g>",
                        )
                        await publish(ConfigReload("basic", key, new_basic[key], old_basic[key]))
                for key in set(new_basic) - set(old_basic):
                    logger.debug(f"Basic config <y>{key!r}</y> appended")
                    await publish(ConfigReload("basic", key, new_basic[key]))
                for plugin_name in old_plugin:
                    if plugin_name.startswith("$"):
                        continue
                    pid = plugin_name.replace("::", "arclet.entari.builtins.")
                    if plugin_name not in EntariConfig.instance.plugin:
                        if plugin := find_plugin(pid):
                            if plugin.is_static:
                                logger.info(f"Plugin <y>{plugin.id!r}</y> is static, ignored.")
                            else:
                                del plugin
                                await unload_plugin_async(pid)
                                logger.info(f"Disposed plugin <blue>{pid!r}</blue>")
                        continue
                    old_conf = EntariConfig._clean(old_plugin[plugin_name])
                    new_conf = EntariConfig.instance.plugin[plugin_name]
                    if old_conf != new_conf:
                        if plugin := find_plugin(pid):
                            added = set(new_conf) - set(old_conf)
                            removed = set(old_conf) - set(new_conf)
                            changed = {k for k in set(new_conf) & set(old_conf) if new_conf[k] != old_conf[k]}
                            changes = added | removed | changed
                            if "$disable" in changes:
                                plugin.disable() if new_conf.get("$disable", False) else plugin.enable()
                                changes.remove("$disable")
                            if "$dry" in changes:
                                changes.remove("$dry")
                                if new_conf.get("$dry", False):
                                    logger.debug(f"Plugin <y>{plugin.id!r}</y> is dry, ignored.")
                                    continue
                            if not changes:
                                continue
                            logger.debug(
                                f"Plugin <y>{plugin_name!r}</y> config changed from <r>{old_conf!r}</r> "
                                f"to <g>{new_conf!r}</g>",
                            )
                            res = await post(
                                ConfigReload("plugin", plugin_name, new_conf, old_conf),
                            )
                            if res and res.value:
                                logger.debug(f"Plugin <y>{pid!r}</y> config change handled by itself.")
                                continue
                            if plugin.is_static:
                                logger.info(f"Plugin <y>{plugin.id!r}</y> is static, ignored.")
                                continue
                            logger.info(f"Detected config of <blue>{pid!r}</blue> changed, reloading...")
                            plugin_file = str(plugin.module.__file__)
                            _conf = plugin.config.copy()

                            async def _():
                                await unload_plugin_async(pid)
                                if plugin := load_plugin(plugin_name, new_conf):
                                    logger.info(f"Reloaded <blue>{plugin.id!r}</blue>")
                                    del plugin
                                else:
                                    logger.error(f"Failed to reload <blue>{plugin_name!r}</blue>")
                                    self.fail[plugin_file] = (pid, _conf)

                            await asyncio.shield(_())
                        else:
                            logger.info(f"Detected <blue>{pid!r}</blue> appended, loading...")
                            load_plugin(plugin_name, new_conf)
                if new := (set(EntariConfig.instance.plugin) - set(old_plugin)):
                    for plugin_name in new:
                        if plugin_name.startswith("$") or plugin_name.startswith("~"):
                            continue
                        if not (plugin := load_plugin(plugin_name)):
                            continue
                        del plugin

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            logger.info(f"Watching directories: {', '.join(repr(d) for d in self.config.watch_dirs)}")
            if self.config.watch_config:
                logger.info("Watching config file changes")
        async with self.stage("blocking"):
            watch_task = asyncio.create_task(self.watch())
            sigexit_task = asyncio.create_task(manager.status.wait_for_sigexit())
            watch_config_task = asyncio.create_task(self.watch_config())
            done, pending = await any_completed(
                sigexit_task,
                watch_task,
                watch_config_task,
            )
            if sigexit_task in done:
                watch_task.cancel()
                watch_config_task.cancel()
        async with self.stage("cleanup"):
            self.fail.clear()


conf = plugin_config(Config)
add_service(Watcher(conf))



"""dataclass schema generator

Modified from https://github.com/Peter554/dc_schema.
Source code is available at https://github.com/GraiaProject/kayaku/blob/master/kayaku/schema_gen.py.


MIT License

Copyright (c) 2022 Peter Byfield

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import dataclasses
import datetime
import enum
import inspect
import numbers
import pathlib
import re
import typing as t
import typing_extensions as t_e

from tarina.generic import Unions, is_typed_dict

from .util import store_field_description

_MISSING = dataclasses.MISSING


class EmptyTypedDict(t.TypedDict):
    pass


@t.runtime_checkable
class DataClass(t.Protocol):
    __dataclass_fields__: t.ClassVar[dict[str, dataclasses.Field]]


def is_sub_type(sub: t.Any, parent: t.Any) -> bool:
    sub_origin = t_e.get_origin(sub) or sub
    return isinstance(sub_origin, type) and issubclass(sub_origin, parent) or sub_origin == parent


# fmt: off
_Format = t.Literal["date-time", "time", "date", "duration", "email", "idn-email", "hostname", "idn-hostname", "ipv4", "ipv6", "uuid", "uri", "uri-reference", "iri", "iri-reference", "regex"]  # noqa: E501
# fmt: on

SCHEMA_ANNO_KEY_MAP = {
    "min_length": "minLength",
    "max_length": "maxLength",
    "exclusive_minimum": "exclusiveMinimum",
    "exclusive_maximum": "exclusiveMaximum",
    "multiple_of": "multipleOf",
    "min_items": "minItems",
    "max_items": "maxItems",
    "unique_items": "uniqueItems",
}


@dataclasses.dataclass(frozen=True)
class Schema:
    title: str | None = None
    description: str | None = None
    examples: list[t.Any] | None = None
    deprecated: bool | None = None

    def schema(self):
        return {SCHEMA_ANNO_KEY_MAP.get(k, k): v for k, v in dataclasses.asdict(self).items() if v is not None}


@dataclasses.dataclass(frozen=True)
class StringSchema(Schema):
    min_length: int | None = None
    max_length: int | None = None
    pattern: str | None = None
    format: _Format | None = None


@dataclasses.dataclass(frozen=True)
class NumberSchema(Schema):
    minimum: numbers.Number | None = None
    maximum: numbers.Number | None = None
    exclusive_minimum: numbers.Number | None = None
    exclusive_maximum: numbers.Number | None = None
    multiple_of: numbers.Number | None = None


@dataclasses.dataclass(frozen=True)
class ContainerSchema(Schema):
    min_items: int | None = None
    max_items: int | None = None
    unique_items: bool | None = None


class SchemaGenerator:
    def __init__(self, dc: type[DataClass] | None = None, ref_root: str = "/") -> None:
        self.root = dc
        self.seen_root = False
        self.ref_root = ref_root
        self.defs = {}

    def format_docstring_description(self, field: dataclasses.Field, description: str) -> str | None:
        return description

    @classmethod
    def from_dc(cls, dc: type[DataClass], ref_root: str = "/") -> dict[str, t.Any]:
        generator = cls(dc, ref_root)
        schema = generator.get_dc_schema(dc)
        if generator.defs:
            schema["$defs"] = generator.defs

        return schema

    def get_dc_schema(self, dc: type[DataClass]) -> dict[str, t.Any]:
        if dc == self.root:
            if self.seen_root:
                return {"$ref": f"#{self.ref_root}/"}
            self.seen_root = True
            return self.create_dc_schema(dc)
        else:
            name = dc.__qualname__
            if name not in self.defs:
                schema = self.create_dc_schema(dc)
                self.defs[name] = schema
            return {"$ref": f"#{self.ref_root}$defs/{name}"}

    def create_dc_schema(self, dc: type[DataClass]):
        schema = {
            "type": "object",
            "title": dc.__qualname__,
            "properties": {},
            "required": [],
            "additionalProperties": False,
        }
        store_field_description(dc, dc.__dataclass_fields__)
        type_hints = t.get_type_hints(dc, include_extras=True)
        if (
            dc.__doc__ is not None
            and dc.__doc__
            != f"{dc.__name__}{str(inspect.signature(dc)).replace(' -> None', '')}"  # Ignore the generated __doc__
        ):
            schema["description"] = dc.__doc__
        for field in dataclasses.fields(dc):
            typ: t.Any = type_hints[field.name]
            field.type = typ  # type: ignore[assignment]
            default_ = field.default_factory() if field.default_factory is not _MISSING else field.default
            if (f_description := field.metadata.get("description")) is not None:
                f_a: Schema
                base, f_a = t.get_args(typ) if t.get_origin(typ) == t.Annotated else (typ, Schema())
                if f_a.description is None:
                    object.__setattr__(
                        f_a,
                        "description",
                        self.format_docstring_description(field, f_description),
                    )
                typ = t.Annotated[base, f_a]
            schema["properties"][field.name] = self.get_field_schema(typ, default_)
            schema["properties"][field.name]["title"] = field.name.title().replace("_", " ")
            field_is_optional = field.default is not _MISSING or field.default_factory is not _MISSING
            if not field_is_optional:
                schema["required"].append(field.name)
        if not schema["required"]:
            schema.pop("required")
        return schema

    def get_simple_schema(self, typ: type, default: t.Any):
        if typ is t.Any:
            return self.get_any_schema(default)
        elif typ in (None, type(None)):
            return self.get_none_schema(default)
        elif is_sub_type(typ, str):
            return self.get_str_schema(default)
        elif is_sub_type(typ, bool):
            return self.get_bool_schema(default)
        elif is_sub_type(typ, int):
            return self.get_int_schema(default)
        elif is_sub_type(typ, numbers.Number):
            return self.get_number_schema(default)
        elif is_sub_type(typ, enum.Enum):
            return self.get_enum_schema(typ, default)
        elif is_sub_type(typ, datetime.datetime):
            return self.get_datetime_schema()
        elif is_sub_type(typ, datetime.date):
            return self.get_date_schema()
        elif is_sub_type(typ, re.Pattern):
            return self.get_regex_schema()
        elif is_sub_type(typ, pathlib.Path):
            return self.get_str_schema(default)

    def get_complex_schema(self, typ: type, default: t.Any):
        if dataclasses.is_dataclass(typ):
            return self.get_dc_schema(typ)
        elif t.get_origin(typ) in Unions:
            return self.get_union_schema(typ, default)
        elif t.get_origin(typ) == t.Literal:
            return self.get_literal_schema(typ, default)
        elif t.get_origin(typ) == t.Annotated:
            return self.get_annotated_schema(typ, default)
        elif is_typed_dict(typ) or t_e.is_typeddict(typ):
            return self.get_typed_dict_schema(typ)
        elif is_sub_type(typ, dict):
            return self.get_dict_schema(typ)
        elif is_sub_type(typ, list):
            return self.get_list_schema(typ)
        elif is_sub_type(typ, tuple):
            return self.get_tuple_schema(typ, default)
        elif is_sub_type(typ, set):
            return self.get_set_schema(typ)

    def get_field_schema(self, typ: type, default: t.Any):
        if (schema := self.get_simple_schema(typ, default)) is not None:
            return schema
        if (schema := self.get_complex_schema(typ, default)) is not None:
            return schema
        raise NotImplementedError(f"field type '{typ}' not implemented")

    def get_any_schema(self, default: t.Any):
        return {} if default is _MISSING else {"default": default}

    def get_union_schema(self, typ: type, default: t.Any):
        args = t.get_args(typ)
        if str in args and pathlib.Path in args:
            args = tuple(i for i in args if i is not pathlib.Path)
        if len(args) == 1:
            return self.get_field_schema(args[0], default)
        if len(args) == 2 and type(None) in args and default is not _MISSING:
            return self.get_field_schema(
                args[0] if args[1] is type(None) else args[1], _MISSING if default is None else default
            )

        schemas = [self.get_field_schema(arg, _MISSING) for arg in args]
        # 需要确定是使用 anyOf 还是 oneOf
        # anyOf 允许多个 schema 同时匹配
        # oneOf 则要求只能有一个 schema 匹配
        # 理论上来说，anyOf 更宽松一些，但在某些情况下可能，例如 type 不一致
        types = {}
        for sch in schemas:
            types.setdefault(sch.get("type"), []).append(sch)

        if len(types) == 1:
            return {"oneOf": schemas, **({"default": default} if default is not _MISSING else {})}
        return {
            "oneOf": [
                schs[0] if len(schs) == 1 else {"oneOf": schs} if typ == "object" else {"anyOf": schs}
                for typ, schs in types.items()
            ],
            **({"default": default} if default is not _MISSING else {}),
        }

    def get_literal_schema(self, typ, default):
        schema = {} if default is _MISSING else {"default": default}
        args = t.get_args(typ)
        if len(args) == 1:
            return {**schema, "const": args[0]}
        return {"enum": list(args), **schema}

    def get_dict_schema(self, typ):
        args = t.get_args(typ)
        assert len(args) in {0, 2}
        if args:
            return {"type": "object", "additionalProperties": self.get_field_schema(args[1], _MISSING)}
        else:
            return {"type": "object"}

    # fmt: off
    def get_typed_dict_schema(self, typ: type[EmptyTypedDict]):
        fields: list[tuple[str, t.Any]] = []
        required: list[str] = []  # Python 3.8- don't have `__required_keys__`
        for name, anno in t_e.get_type_hints(typ, include_extras=True).items():
            anno: t.Any
            schema_anno: Schema | None = None
            if t_e.get_origin(anno) == t_e.Annotated:
                anno, schema_anno = t_e.get_args(anno)
            if t_e.get_origin(anno) == t_e.Required or typ.__total__ and t_e.get_origin(anno) != t_e.NotRequired:
                required.append(name)
            if t_e.get_origin(anno) in (t_e.Required, t_e.NotRequired):
                anno = t_e.get_args(anno)[0]
            fields.append((name, (anno if schema_anno is None else t.Annotated[anno, schema_anno])))
        dc: type[DataClass] = t.cast(type[DataClass], dataclasses.make_dataclass(typ.__qualname__, fields))
        dc.__module__ = typ.__module__
        store_field_description(typ, dc.__dataclass_fields__)
        dc_schema = self.get_dc_schema(dc)
        dc_def = self.defs[dc.__qualname__]
        dc_def["required"] = required
        return dc_schema

    def get_list_schema(self, typ):
        args = t.get_args(typ)
        assert len(args) in {0, 1}
        if args:
            return {"type": "array", "items": self.get_field_schema(args[0], _MISSING)}
        else:
            return {"type": "array"}

    def get_tuple_schema(self, typ, default):
        schema = {} if default is _MISSING else {"default": list(default)}
        args = t.get_args(typ)
        if args and len(args) == 2 and args[1] is ...:
            schema = {"type": "array", "items": self.get_field_schema(args[0], _MISSING), **schema}
        elif args:
            schema = {"type": "array", "prefixItems": [self.get_field_schema(arg, _MISSING) for arg in args], "minItems": len(args), "maxItems": len(args), **schema}  # noqa: E501
        else:
            schema = {"type": "array", **schema}
        return schema

    def get_set_schema(self, typ):
        args = t.get_args(typ)
        assert len(args) in {0, 1}
        if args:
            return {"type": "array", "items": self.get_field_schema(args[0], _MISSING), "uniqueItems": True}
        else:
            return {"type": "array", "uniqueItems": True}
    # fmt: on
    def get_none_schema(self, default):
        if default is _MISSING:
            return {"type": "null"}
        else:
            return {"type": "null", "default": default}

    def get_str_schema(self, default):
        if default is _MISSING:
            return {"type": "string"}
        else:
            return {"type": "string", "default": default}

    def get_bool_schema(self, default):
        if default is _MISSING:
            return {"type": "boolean"}
        else:
            return {"type": "boolean", "default": default}

    def get_int_schema(self, default):
        if default is _MISSING:
            return {"type": "integer"}
        else:
            return {"type": "integer", "default": default}

    def get_number_schema(self, default):

        if default is _MISSING:
            return {"type": "number"}
        else:
            return {"type": "number", "default": default}

    def get_enum_schema(self, typ: type[enum.Enum], default: enum.Enum):
        name = typ.__qualname__
        title = typ.__qualname__.title()
        if name not in self.defs:
            self.defs[name] = {"title": title, "enum": [v.value for v in typ]}
        if default is _MISSING:
            return {"$ref": f"#{self.ref_root}$defs/{name}"}
        else:
            return {"$ref": f"#{self.ref_root}$defs/{name}", "default": default.value}

    # fmt: off
    def get_annotated_schema(self, typ, default):
        args = t.get_args(typ)
        schema = self.get_field_schema(args[0], default)
        for meta in args[1:]:
            if isinstance(meta, Schema):
                if isinstance(meta, StringSchema) and not is_sub_type(args[0], str):
                    raise TypeError(f"Trying to apply string-specific annotation to {args[0]}")
                elif isinstance(meta, ContainerSchema) and not any(is_sub_type(args[0], typ) for typ in (list, tuple, set)):  # noqa: E501
                    raise TypeError(f"Trying to apply sequence-specific annotation to {args[0]}")
                elif isinstance(meta, NumberSchema) and not is_sub_type(args[0], numbers.Number) and args[0] is bool:  # noqa: E501
                    raise TypeError(f"Trying to apply number-specific annotation to {args[0]}")
                schema.update(meta.schema())
            elif isinstance(meta, t_e.Doc):
                description = self.format_docstring_description(dataclasses.fields(typ)[0], meta.documentation)
                if description:
                    schema["description"] = description
            elif isinstance(meta, str) and meta in t.get_args(_Format):
                schema["format"] = meta
        return schema
    # fmt: on
    def get_datetime_schema(self):
        return {"type": "string", "format": "date-time"}

    def get_date_schema(self):
        return {"type": "string", "format": "date"}

    def get_regex_schema(self):
        return {"type": "string", "format": "regex"}


def update_schema_ref(root: dict, sect: tuple[str, ...], name: str) -> None:
    for s in sect:
        root["type"] = "object"
        required: list[str] = root.setdefault("required", [])
        if s not in required:
            required.append(s)
        root = root.setdefault("properties", {})
        root = root.setdefault(s, {})

    all_of: list[dict[str, str]] = root.setdefault("allOf", [])
    if {"$ref": f"#/$defs/{name}"} not in all_of:
        all_of.append({"$ref": f"#/$defs/{name}"})



from __future__ import annotations

from collections.abc import Awaitable, Callable, Iterable, Sequence
from copy import deepcopy
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar, Union, overload
from typing_extensions import Self, SupportsIndex

from satori import select as satori_select
from satori.element import At, Element, Link, Sharp, Style, Text, transform
from satori.parser import parse

T = TypeVar("T")
TE = TypeVar("TE", bound=Element)
TE1 = TypeVar("TE1", bound=Element)

Fragment: TypeAlias = Element | Iterable[Element]
Visit: TypeAlias = Callable[[Element], T]
Render: TypeAlias = Callable[[dict[str, Any], list[Element]], T]
SyncTransformer: TypeAlias = bool | Fragment | Render[bool | Fragment]
AsyncTransformer: TypeAlias = bool | Fragment | Render[Awaitable[bool | Fragment]]
SyncVisitor: TypeAlias = dict[str, SyncTransformer] | Visit[bool | Fragment]
AsyncVisitor: TypeAlias = dict[str, AsyncTransformer] | Visit[Awaitable[bool | Fragment]]

MessageContainer = Union[str, Element, Sequence["MessageContainer"], "MessageChain[Element]"]


class MessageChain(list[TE]):
    """消息序列

    Args:
        message: 消息内容
    """

    @overload
    def __init__(self): ...

    @overload
    def __init__(self: MessageChain[Text], message: str): ...

    @overload
    def __init__(self, message: TE): ...

    @overload
    def __init__(self: MessageChain[TE1], message: TE1): ...

    @overload
    def __init__(self, message: Iterable[TE]): ...

    @overload
    def __init__(self: MessageChain[TE1], message: Iterable[TE1]): ...

    @overload
    def __init__(self: MessageChain[Text], message: Iterable[str]): ...

    @overload
    def __init__(self: MessageChain[Text | TE1], message: Iterable[str | TE1]): ...

    def __init__(
        self: MessageChain[Element],
        message: Iterable[str | TE] | str | TE | None = None,
    ):
        super().__init__()
        if isinstance(message, str):
            self.__iadd__(Text(message))
        elif isinstance(message, Iterable):
            for i in message:
                self.__iadd__(Text(i) if isinstance(i, str) else i)
        elif isinstance(message, Element):
            self.__iadd__(message)

    def __str__(self) -> str:
        return "".join(str(elem) for elem in self)

    def __repr__(self) -> str:
        return "[" + ", ".join(repr(elem) for elem in self) + "]"

    @overload
    def __add__(self, other: str) -> MessageChain[TE | Text]: ...

    @overload
    def __add__(self, other: TE | Iterable[TE]) -> MessageChain[TE]: ...

    @overload
    def __add__(self, other: TE1 | Iterable[TE1]) -> MessageChain[TE | TE1]: ...

    def __add__(self, other: str | TE | TE1 | Iterable[TE | TE1]) -> MessageChain:
        result: MessageChain = self.copy()
        if isinstance(other, str):
            if result and isinstance(text := result[-1], Text):
                result[-1] = Text(text.text + other)
            else:
                result.append(Text(other))
        elif isinstance(other, Element):
            if result and isinstance(result[-1], Text) and isinstance(other, Text):
                result[-1] = Text(result[-1].text + other.text)
            else:
                result.append(other)
        elif isinstance(other, Iterable):
            for elem in other:
                result += elem
        else:
            raise TypeError(f"Unsupported type {type(other)!r}")
        return result

    @overload
    def __radd__(self, other: str) -> MessageChain[Text | TE]: ...

    @overload
    def __radd__(self, other: TE | Iterable[TE]) -> MessageChain[TE]: ...

    @overload
    def __radd__(self, other: TE1 | Iterable[TE1]) -> MessageChain[TE1 | TE]: ...

    def __radd__(self, other: str | TE1 | Iterable[TE1]) -> MessageChain:
        result = MessageChain(other)
        return result + self

    def __iadd__(self, other: str | TE | Iterable[TE]) -> Self:
        if isinstance(other, str):
            if self and isinstance(text := self[-1], Text):
                list.__setitem__(self, -1, Text(text.text + other))
            else:
                self.append(Text(other))  # type: ignore
        elif isinstance(other, Element):
            if self and (isinstance(text := self[-1], Text) and isinstance(other, Text)):
                list.__setitem__(self, -1, Text(text.text + other.text))
            else:
                self.append(other)
        elif isinstance(other, Iterable):
            for elem in other:
                self.__iadd__(elem)
        else:
            raise TypeError(f"Unsupported type {type(other)!r}")
        return self

    @overload
    def __getitem__(self, args: type[TE1]) -> MessageChain[TE1]:
        """获取仅包含指定消息段类型的消息

        Args:
            args: 消息段类型

        Returns:
            所有类型为 `args` 的消息段
        """

    @overload
    def __getitem__(self, args: tuple[type[TE1], int]) -> TE1:
        """索引指定类型的消息段

        Args:
            args: 消息段类型和索引

        Returns:
            类型为 `args[0]` 的消息段第 `args[1]` 个
        """

    @overload
    def __getitem__(self, args: tuple[type[TE1], slice]) -> MessageChain[TE1]:
        """切片指定类型的消息段

        Args:
            args: 消息段类型和切片

        Returns:
            类型为 `args[0]` 的消息段切片 `args[1]`
        """

    @overload
    def __getitem__(self, args: int) -> TE:
        """索引消息段

        Args:
            args: 索引

        Returns:
            第 `args` 个消息段
        """

    @overload
    def __getitem__(self, args: slice) -> Self:
        """切片消息段

        Args:
            args: 切片

        Returns:
            消息切片 `args`
        """

    def __getitem__(
        self,
        args: type[TE1] | tuple[type[TE1], int] | tuple[type[TE1], slice] | int | slice,
    ) -> TE | TE1 | MessageChain[TE1] | Self:
        arg1, arg2 = args if isinstance(args, tuple) else (args, None)
        if isinstance(arg1, int) and arg2 is None:
            return super().__getitem__(arg1)
        if isinstance(arg1, slice) and arg2 is None:
            return MessageChain(super().__getitem__(arg1))  # type: ignore
        if TYPE_CHECKING:
            assert not isinstance(arg1, slice | int)
        if issubclass(arg1, Element) and arg2 is None:
            return MessageChain(elem for elem in self if isinstance(elem, arg1))  # type: ignore
        if issubclass(arg1, Element) and isinstance(arg2, int):
            return [elem for elem in self if isinstance(elem, arg1)][arg2]
        if issubclass(arg1, Element) and isinstance(arg2, slice):
            return MessageChain([elem for elem in self if isinstance(elem, arg1)][arg2])  # type: ignore
        raise ValueError("Incorrect arguments to slice")  # pragma: no cover

    def __contains__(self, value: str | Element | type[Element]) -> bool:
        """检查消息段是否存在

        Args:
            value: 消息段或消息段类型
        Returns:
            消息内是否存在给定消息段或给定类型的消息段
        """
        if isinstance(value, type):
            return bool(next((elem for elem in self if isinstance(elem, value)), None))
        if isinstance(value, str):
            value = Text(value)
        return super().__contains__(value)

    def has(self, value: str | Element | type[Element]) -> bool:
        return value in self

    def index(self, value: str | Element | type[Element], *args: SupportsIndex) -> int:
        """索引消息段

        Args:
            value: 消息段或者消息段类型
            args: start 与 end

        Returns:
            索引 index

        Raise:
            ValueError: 消息段不存在
        """
        if isinstance(value, type):
            first_elemment = next((elem for elem in self if isinstance(elem, value)), None)
            if first_elemment is None:
                raise ValueError(f"Element with type {value!r} is not in message")
            return super().index(first_elemment, *args)
        if isinstance(value, str):
            value = Text(value)
        return super().index(value, *args)  # type: ignore

    def get(self, type_: type[TE], count: int | None = None) -> MessageChain[TE]:
        """获取指定类型的消息段

        Args:
            type_: 消息段类型
            count: 获取个数

        Returns:
            构建的新消息
        """
        if count is None:
            return self[type_]

        iterator, filtered = (elem for elem in self if isinstance(elem, type_)), MessageChain()
        for _ in range(count):
            elem = next(iterator, None)
            if elem is None:
                break
            filtered.append(elem)
        return filtered  # type: ignore

    def count(self, value: type[Element] | str | Element) -> int:
        """计算指定消息段的个数

        Args:
            value: 消息段或消息段类型

        Returns:
            个数
        """
        if isinstance(value, str):
            value = Text(value)
        return (
            len(self[value])  # type: ignore
            if isinstance(value, type)
            else super().count(value)  # type: ignore
        )

    def only(self, value: type[Element] | str | Element) -> bool:
        """检查消息中是否仅包含指定消息段

        Args:
            value: 指定消息段或消息段类型

        Returns:
            是否仅包含指定消息段
        """
        if isinstance(value, type):
            return all(isinstance(elem, value) for elem in self)
        if isinstance(value, str):
            value = Text(value)
        return all(elem == value for elem in self)

    def join(self, iterable: Iterable[TE1 | MessageChain[TE1]]) -> MessageChain[TE | TE1]:
        """将多个消息连接并将自身作为分割

        Args:
            iterable: 要连接的消息

        Returns:
            连接后的消息
        """
        ret = MessageChain()
        for index, msg in enumerate(iterable):
            if index != 0:
                ret.extend(self)
            if isinstance(msg, Element):
                ret.append(msg)
            else:
                ret.extend(msg.copy())
        return ret  # type: ignore

    def copy(self) -> MessageChain[TE]:
        """深拷贝消息"""
        return deepcopy(self)

    def fork(self) -> MessageChain[TE]:
        """浅拷贝消息"""
        new = self.__class__()
        list.extend(new, self)
        return new

    def include(self, *types: type[Element]) -> MessageChain:
        """过滤消息

        Args:
            types: 包含的消息段类型

        Returns:
            新构造的消息
        """
        return MessageChain(elem for elem in self if elem.__class__ in types)

    def exclude(self, *types: type[Element]) -> MessageChain:
        """过滤消息

        Args:
            types: 不包含的消息段类型

        Returns:
            新构造的消息
        """
        return MessageChain(elem for elem in self if elem.__class__ not in types)

    def extract_plain_text(self) -> str:
        """提取消息内纯文本消息"""

        return "".join(elem.text for elem in self if isinstance(elem, Text))

    def filter(self, predicate: Callable[[TE], bool]) -> MessageChain[TE]:
        """过滤消息

        Args:
            predicate: 过滤函数
        """
        return MessageChain(elem for elem in self if predicate(elem))

    @overload
    def map(self, func: Callable[[TE], TE1]) -> MessageChain[TE1]: ...

    @overload
    def map(self, func: Callable[[TE], T]) -> list[T]: ...

    def map(self, func: Callable[[TE], TE1] | Callable[[TE], T]) -> MessageChain[TE1] | list[T]:
        result1 = []
        result2 = []
        for elem in self:
            result = func(elem)
            if isinstance(result, Element):
                result1.append(result)
            else:
                result2.append(result)
        if result1:
            return MessageChain(result1)
        return result2

    def select(self, cls: type[TE1]) -> MessageChain[TE1]:
        return MessageChain(satori_select(list(self), cls))

    @staticmethod
    def _visit_sync(elem: Element, rules: SyncVisitor):
        _type, data, children = elem.tag, elem._attrs, elem.children
        if not isinstance(rules, dict):
            return rules(elem)
        result = rules.get(_type, True)
        if not isinstance(result, bool | Element | Iterable):
            result = result(data, children)
        return result

    @staticmethod
    async def _visit_async(elem: Element, rules: AsyncVisitor):
        _type, data, children = elem.tag, elem._attrs, elem.children
        if not isinstance(rules, dict):
            return await rules(elem)
        result = rules.get(_type, True)
        if not isinstance(result, bool | Element | Iterable):
            result = await result(data, children)
        return result

    def transform(self, rules: SyncVisitor) -> MessageChain:
        """同步遍历消息元素并转换

        Args:
            rules: 转换规则

        Returns:
            转换后的消息
        """
        output = MessageChain()
        for elem in self:
            result = self._visit_sync(elem, rules)
            if result is True:
                output += elem
            elif result is not False:
                if isinstance(result, Element):
                    output += result
                else:
                    output.extend(result)
        return output

    async def transform_async(self, rules: AsyncVisitor) -> MessageChain:
        """异步遍历消息段并转换

        Args:
            rules: 转换规则

        Returns:
            转换后的消息
        """
        output = MessageChain()
        for elem in self:
            result = await self._visit_async(elem, rules)
            if result is True:
                output += elem
            elif result is not False:
                if isinstance(result, Element):
                    output += result
                else:
                    output.extend(result)
        return output

    def split(self, pattern: str = " ") -> list[Self]:
        """和 `str.split` 差不多, 提供一个字符串, 然后返回分割结果.

        Args:
            pattern (str): 分隔符. 默认为单个空格.

        Returns:
            list[Self]: 分割结果, 行为和 `str.split` 差不多.
        """

        result: list[Self] = []
        tmp = []
        for seg in self:
            if isinstance(seg, Text):
                split_result = seg.text.split(pattern)
                for index, split_text in enumerate(split_result):
                    if tmp and index > 0:
                        result.append(self.__class__(tmp))
                        tmp = []
                    if split_text:
                        tmp.append(split_text)
            else:
                tmp.append(seg)
        if tmp:
            result.append(self.__class__(tmp))
            tmp = []
        return result

    def replace(
        self,
        old: str,
        new: str,
    ) -> Self:
        """替换消息中有关的文本

        Args:
            old (str): 要替换的字符串.
            new (str): 替换后的字符串.

        Returns:
            UniMessage: 修改后的消息链, 若未替换则原样返回.
        """
        result_list: list[TE] = []
        for seg in self:
            if isinstance(seg, Text):
                result_list.append(seg.__class__(seg.text.replace(old, new)))
            else:
                result_list.append(seg)
        return self.__class__(result_list)

    def startswith(self, string: str) -> bool:
        """判断消息链是否以给出的字符串开头

        Args:
            string (str): 字符串

        Returns:
            bool: 是否以给出的字符串开头
        """

        if not self or not isinstance(self[0], Text):
            return False
        return list.__getitem__(self, 0).text.startswith(string)

    def endswith(self, string: str) -> bool:
        """判断消息链是否以给出的字符串结尾

        Args:
            string (str): 字符串

        Returns:
            bool: 是否以给出的字符串结尾
        """

        if not self or not isinstance(self[-1], Text):
            return False
        return list.__getitem__(self, -1).text.endswith(string)

    def removeprefix(self, prefix: str) -> Self:
        """移除消息链前缀.

        Args:
            prefix (str): 要移除的前缀.

        Returns:
            UniMessage: 修改后的消息链.
        """
        copy = list.copy(self)
        if not copy:
            return self.__class__(copy)
        seg = copy[0]
        if not isinstance(seg, Text):
            return self.__class__(copy)
        if seg.text.startswith(prefix):
            seg = seg.__class__(seg.text[len(prefix) :])
            if not seg.text:
                copy.pop(0)
            else:
                copy[0] = seg
        return self.__class__(copy)

    def removesuffix(self, suffix: str) -> Self:
        """移除消息链后缀.

        Args:
            suffix (str): 要移除的后缀.

        Returns:
            UniMessage: 修改后的消息链.
        """
        copy = list.copy(self)
        if not copy:
            return self.__class__(copy)
        seg = copy[-1]
        if not isinstance(seg, Text):
            return self.__class__(copy)
        if seg.text.endswith(suffix):
            seg = seg.__class__(seg.text[: -len(suffix)])
            if not seg.text:
                copy.pop(-1)
            else:
                copy[-1] = seg
        return self.__class__(copy)

    def strip(self, *segments: str | Element | type[Element]) -> Self:
        return self.lstrip(*segments).rstrip(*segments)

    def lstrip(self, *segments: str | Element | type[Element]) -> Self:
        types = [i for i in segments if not isinstance(i, str)] or []
        chars = "".join([i for i in segments if isinstance(i, str)]) or None
        copy = list.copy(self)
        if not copy:
            return self.__class__(copy)
        while copy:
            seg = copy[0]
            if seg in types or seg.__class__ in types:
                copy.pop(0)
            elif isinstance(seg, Text):
                seg = seg.__class__(seg.text.lstrip(chars))
                if not seg.text:
                    copy.pop(0)
                    continue
                else:
                    copy[0] = seg
                break
            else:
                break
        return self.__class__(copy)

    def rstrip(self, *segments: str | Element | type[Element]) -> Self:
        types = [i for i in segments if not isinstance(i, str)] or []
        chars = "".join([i for i in segments if isinstance(i, str)]) or None
        copy = list.copy(self)
        if not copy:
            return self.__class__(copy)
        while copy:
            seg = copy[-1]
            if seg in types or seg.__class__ in types:
                copy.pop(-1)
            elif isinstance(seg, Text):
                seg = seg.__class__(seg.text.rstrip(chars))
                if not seg.text:
                    copy.pop(-1)
                    continue
                else:
                    copy[-1] = seg
                break
            else:
                break
        return self.__class__(copy)

    def display(self):
        return "".join(
            str(elem) if isinstance(elem, (Text, Style, At, Sharp, Link)) else elem.__class__.__name__  # noqa: UP038
            for elem in self
        )

    @staticmethod
    def of(text: str) -> MessageChain:
        return MessageChain(transform(parse(text)))



from .file import BasicConfModel as BasicConfModel
from .file import EntariConfig as EntariConfig
from .file import load_config as load_config
from .file import model_field as model_field
from .model import ConfigModelAction as ConfigModelAction
from .model import config_model_dump as config_model_dump
from .model import config_model_keys as config_model_keys
from .model import config_model_schema as config_model_schema
from .model import config_model_validate as config_model_validate



from arclet.alconna import Alconna, Args, CommandMeta
from satori.element import At, Author, Sharp, select

from arclet.entari import MessageEvent, Session, command, metadata
from arclet.entari.command.model import Match

from .i18n import Lang

metadata(
    "Inspect",
    [{"name": "RF-Tar-Railt", "email": "rf_tar_railt@qq.com"}],
    description="Inspect on any user, group or channel",
    readme="README.md",
)


inspect_cmd = Alconna(
    "inspect",
    Args["target?", [At, Sharp]],
    meta=CommandMeta(Lang.entari_plugin_inspect.description.cast(), example="inspect @user\ninspect #channel\ninspect"),
)


SceneNames = {
    "DIRECT": Lang.entari_plugin_inspect.scene.direct,
    "TEXT": Lang.entari_plugin_inspect.scene.text,
    "VOICE": Lang.entari_plugin_inspect.scene.voice,
    "CATEGORY": Lang.entari_plugin_inspect.scene.category,
}
TP = "{platform}\n{self}\n{scene}\n{guild}\n{channel}\n{user}"


@command.on(inspect_cmd)
async def inspect(session: Session[MessageEvent], target: Match["At | Sharp"]):
    event = session.event
    texts = {
        "platform": Lang.entari_plugin_inspect.platform(platform=session.account.platform),
        "self": Lang.entari_plugin_inspect.self(self_id=session.account.self_id),
        "scene": Lang.entari_plugin_inspect.scene.name(scene=SceneNames[session.channel.type.name]()),
        "guild": Lang.entari_plugin_inspect.guild(guild_id=event.guild.id if event.guild else None),
        "channel": Lang.entari_plugin_inspect.channel(channel_id=session.channel.id),
    }
    if target.available:
        if isinstance(target.result, At) and target.result.id:
            texts["user"] = Lang.entari_plugin_inspect.user(user_id=target.result.id)
        elif isinstance(target.result, Sharp) and target.result.id:
            texts["channel"] = Lang.entari_plugin_inspect.channel(channel_id=target.result.id)
        else:
            await session.send_message(Lang.entari_plugin_inspect.invalid())
            return
    if "user" not in texts:
        if event.quote and (authors := select(event.quote, Author)):
            texts["user"] = Lang.entari_plugin_inspect.user(user_id=authors[0].id)
        else:
            texts["user"] = Lang.entari_plugin_inspect.user(user_id=session.user.id)

    await session.send_message(TP.format(**texts))
    return



from pathlib import Path
from typing import Any

from arclet.entari.config.file import register_dumper, register_loader

try:
    from tomlkit import dumps, loads
except ImportError:
    dumps = None
    loads = None


@register_loader("toml")
def toml_loader(text: str) -> dict[str, Any]:
    """
    Load a TOML file and return its content as a dictionary.
    """
    if loads is None:
        raise RuntimeError("tomlkit is not installed. Please install with `arclet-entari[toml]`")
    return loads(text)


@register_dumper("toml")
def toml_dumper(save_path: Path, origin: dict[str, Any], indent: int = 4):
    """
    Dump a dictionary to a TOML file.
    """
    if dumps is None:
        raise RuntimeError("tomlkit is not installed. Please install with `arclet-entari[toml]`")
    with open(save_path, "w+", encoding="utf-8") as f:
        f.write(dumps(origin))



from collections.abc import Awaitable, Callable
from typing import TypeAlias, Union

from arclet.letoderea import STOP, Propagator

from ..session import Session
from .message import direct_message, notice_me, public_message, reply_me, to_me


def user(*ids: str):
    async def check_user(session: Session):
        return (session.user.id in ids) if ids else True

    return check_user


def channel(*ids: str):
    async def check_channel(session: Session):
        return (session.channel.id in ids) if ids else True

    return check_channel


def guild(*ids: str):
    async def check_guild(session: Session):
        return (session.guild.id in ids) if ids else True

    return check_guild


def account(*ids: str):
    async def check_account(session: Session):
        return (session.account.self_id in ids) if ids else True

    return check_account


def platform(*ids: str):
    async def check_platform(session: Session):
        return (session.account.platform in ids) if ids else True

    return check_platform


_keys = {
    "user": (user, 2),
    "guild": (guild, 3),
    "channel": (channel, 4),
    "self": (account, 1),
    "platform": (platform, 0),
    "direct": (lambda: direct_message, 5),
    "private": (lambda: direct_message, 5),
    "public": (lambda: public_message, 6),
}

_mess_keys = {
    "reply_me": (reply_me, 7),
    "notice_me": (notice_me, 8),
    "to_me": (to_me, 9),
}

_op_keys = {
    "$and": "and",
    "$or": "or",
    "$not": "not",
    "$intersect": "and",
    "$union": "or",
    "$exclude": "not",
}

PATTERNS: TypeAlias = dict[str, Union[list[str], bool, "PATTERNS"]]


class _Filter(Propagator):
    def __init__(
        self,
        steps: list[Callable[[Session], Awaitable[bool]]],
        mess: list[Callable[[bool, bool], bool]],
        ops: list[tuple[str, "_Filter"]],
    ):
        self.steps = steps
        self.mess = mess
        self.ops = ops

    async def check(self, session: Session | None = None, is_reply_me: bool = False, is_notice_me: bool = False):
        res = True
        if session and self.steps:
            res = all([await step(session) for step in self.steps])
        if self.mess:
            res = res and all(mess(is_reply_me, is_notice_me) for mess in self.mess)
        for op, f_ in self.ops:
            if op == "and":
                res = res and (await f_.check(session, is_reply_me, is_notice_me)) is None  # type: ignore
            elif op == "or":
                res = res or (await f_.check(session, is_reply_me, is_notice_me)) is None  # type: ignore
            else:
                res = res and (await f_.check(session, is_reply_me, is_notice_me)) is STOP  # type: ignore
        return None if res else STOP

    def compose(self):
        yield self.check, True, 0


def parse(patterns: PATTERNS):
    step: dict[int, Callable[[Session], Awaitable[bool]]] = {}
    mess: dict[int, Callable[[bool, bool], bool]] = {}
    ops: list[tuple[str, _Filter]] = []

    for key, value in patterns.items():
        if key in _keys:
            step[_keys[key][1]] = _keys[key][0](*map(str, value)) if isinstance(value, list) else _keys[key][0]()
        elif key in _mess_keys:
            if key == "reply_me":
                mess[_mess_keys[key][1]] = lambda is_reply_me, is_notice_me: (
                    True if _mess_keys[key][0](is_reply_me) is None else False
                )
            elif key == "notice_me":
                mess[_mess_keys[key][1]] = lambda is_reply_me, is_notice_me: (
                    True if _mess_keys[key][0](is_notice_me) is None else False
                )
            else:
                mess[_mess_keys[key][1]] = lambda is_reply_me, is_notice_me: (
                    True if _mess_keys[key][0](is_reply_me, is_notice_me) is None else False
                )
        elif key in _op_keys:
            op = _op_keys[key]
            if not isinstance(value, dict):
                raise ValueError(f"Expect a dict for operator {key}")
            ops.append((op, parse(value)))
        else:
            raise ValueError(f"Unknown key: {key}")

    return _Filter(
        steps=[slot[1] for slot in sorted(step.items(), key=lambda x: x[0])],
        mess=[slot[1] for slot in sorted(mess.items(), key=lambda x: x[0])],
        ops=ops,
    )



import asyncio
from collections.abc import Callable
from dataclasses import dataclass, field
from types import ModuleType
from typing import Any, AsyncGenerator, Awaitable, Generator, Generic, Literal, TypedDict, TypeVar, overload
from typing_extensions import NotRequired, Self

from arclet.letoderea import ExitState, Propagator, Provider, ProviderFactory, Scope, Subscriber
from arclet.letoderea.breakpoint import StepOut
from arclet.letoderea.decorate import Check
from arclet.letoderea.provider import TProviders
from arclet.letoderea.publisher import Publisher
from arclet.letoderea.scope import RegisterWrapper
from arclet.letoderea.typing import Resultable, TCallable
from launart import Service
from tarina import ContextModel

current_plugin: ContextModel[Plugin] = ContextModel("_current_plugin")

T = TypeVar("T")
TE = TypeVar("TE", covariant=True)
TE1 = TypeVar("TE1")
TS = TypeVar("TS", bound=Service)
R = TypeVar("R")
R1 = TypeVar("R1")

class PluginDispatcher(Generic[T]):
    publisher: Publisher
    plugin: Plugin
    _event: type
    providers: list[Provider[Any] | ProviderFactory]
    propagators: list[Propagator]

    def __init__(self, plugin: Plugin, event: type, name: str | None = None): ...
    def waiter(
        self, event: Any = None, providers: TProviders | None = None, priority: int = 15, block: bool = False
    ) -> Callable[[Callable[..., R]], StepOut[R]]: ...
    @overload
    def register(
        self,
        func: Callable[..., Generator[T | ExitState | None, None, None]],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
        once: bool = False,
    ) -> Subscriber[Generator[T, None, None]]: ...
    @overload
    def register(
        self,
        func: Callable[..., AsyncGenerator[T | ExitState | None, None]],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
        once: bool = False,
    ) -> Subscriber[AsyncGenerator[T, None]]: ...
    @overload
    def register(
        self,
        func: Callable[..., Awaitable[T | ExitState | None]],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
        once: bool = False,
    ) -> Subscriber[Awaitable[T]]: ...
    @overload
    def register(
        self,
        func: Callable[..., T | ExitState | None],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
        once: bool = False,
    ) -> Subscriber[T]: ...
    @overload
    def register(
        self, *, priority: int = 16, providers: TProviders | None = None, once: bool = False
    ) -> RegisterWrapper[T, None]: ...  # noqa: E501
    @overload
    def once(
        self,
        func: Callable[..., Generator[T | ExitState | None, None, None]],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
    ) -> Subscriber[Generator[T, None, None]]: ...
    @overload
    def once(
        self,
        func: Callable[..., AsyncGenerator[T | ExitState | None, None]],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
    ) -> Subscriber[AsyncGenerator[T, None]]: ...
    @overload
    def once(
        self,
        func: Callable[..., Awaitable[T | ExitState | None]],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
    ) -> Subscriber[Awaitable[T]]: ...
    @overload
    def once(
        self, func: Callable[..., T | ExitState | None], *, priority: int = 16, providers: TProviders | None = None
    ) -> Subscriber[T]: ...
    @overload
    def once(
        self, *, priority: int = 16, providers: TProviders | None = None
    ) -> RegisterWrapper[T, None]: ...  # noqa: E501
    on = register
    handle = register
    __call__ = register

class Author(TypedDict):
    name: str
    """作者名称"""
    email: NotRequired[str]
    """作者邮箱"""

class DependService(TypedDict):
    id: str | type[Service]
    """服务名称"""
    stage: NotRequired[Literal["preparing", "prepared", "blocking"]]
    """服务阶段"""

@dataclass
class PluginMetadata:
    name: str
    author: list[str | Author] = ...
    version: str | None = ...
    license: str | None = ...
    urls: dict[str, str] | None = ...
    description: str | None = ...
    icon: str | None = ...
    readme: str | None = ...
    classifier: list[str] = ...
    requirements: list[str] = ...
    depend_services: list[type[Service] | str | DependService] = ...
    config: Any | None = ...

    def get_config_schema(self) -> dict[str, Any]: ...

@overload
def inject(*services: type[Service] | str | DependService) -> Callable[[TCallable], TCallable]: ...
@overload
def inject(*services: type[Service] | str | DependService, _is_global: Literal[True]) -> Check: ...
@dataclass
class Plugin:
    id: str
    module: ModuleType

    subplugins: set[str] = ...
    config: dict[str, Any] = ...
    is_static: bool = ...
    path: str = ...
    uid: str | None = ...
    _metadata: PluginMetadata | None = ...
    _is_disposed: bool = ...
    _services: dict[str, Service] = field(init=False, default_factory=dict)
    _dispose_callbacks: list[Callable[[], None]] = field(init=False, default_factory=list)
    _config_key: str = field(init=False)
    _scope: Scope = field(init=False)
    _extra: dict[str, Any] = field(default_factory=dict, init=False)  # extra metadata for inspection
    _apply: Callable[[Plugin], Any] | None = field(default=None, init=False)

    @property
    def reusable(self) -> bool: ...
    @property
    def available(self) -> bool: ...
    @staticmethod
    def current() -> Plugin: ...
    @property
    def metadata(self) -> PluginMetadata | None: ...
    @metadata.setter
    def metadata(self, value: PluginMetadata): ...
    def exec_apply(self) -> None: ...
    @property
    def is_available(self) -> bool: ...
    def enable(self) -> set[asyncio.Task] | None: ...
    def disable(self) -> set[asyncio.Task]: ...
    def collect(self, *disposes: Callable[[], None]) -> Self: ...
    def restore(self) -> None: ...
    def _clean_service(self) -> asyncio.Task: ...
    def dispose(self, *, is_cleanup: bool = False) -> set[asyncio.Task]: ...
    @overload
    def dispatch(self, event: type[Resultable[T]], name: str | None = None) -> PluginDispatcher[T]: ...
    @overload
    def dispatch(self, event: type[Any], name: str | None = None) -> PluginDispatcher[Any]: ...
    @overload
    def use(
        self,
        pub: Publisher[Resultable[T]],
        func: Callable[..., Generator[T | ExitState | None, None, None]],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
    ) -> Subscriber[Generator[T, None, None]]: ...
    @overload
    def use(
        self,
        pub: Publisher[Resultable[T]],
        func: Callable[..., AsyncGenerator[T | ExitState | None, None]],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
    ) -> Subscriber[AsyncGenerator[T, None]]: ...
    @overload
    def use(
        self,
        pub: Publisher[Resultable[T]],
        func: Callable[..., Awaitable[T | ExitState | None]],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
    ) -> Subscriber[Awaitable[T]]: ...
    @overload
    def use(
        self,
        pub: Publisher[Resultable[T]],
        func: Callable[..., T | ExitState | None],
        *,
        priority: int = 16,
        providers: TProviders | None = None,
    ) -> Subscriber[T]: ...
    @overload
    def use(
        self, pub: Publisher[Resultable[T]], *, priority: int = 16, providers: TProviders | None = None
    ) -> RegisterWrapper[T, None]: ...
    @overload
    def use(
        self, pub: Publisher[Any], func: Callable[..., T], *, priority: int = 16, providers: TProviders | None = None
    ) -> Subscriber[T]: ...
    @overload
    def use(
        self, pub: Publisher[Any], *, priority: int = 16, providers: TProviders | None = None
    ) -> RegisterWrapper[None, Callable]: ...
    @overload
    def use(
        self, pub: str, func: Callable[..., T], *, priority: int = 16, providers: TProviders | None = None
    ) -> Subscriber[T]: ...
    @overload
    def use(
        self, pub: str, *, priority: int = 16, providers: TProviders | None = None
    ) -> RegisterWrapper[None, Callable]: ...
    def validate(self, func) -> None: ...
    def proxy(self) -> ModuleType: ...
    def subproxy(self, sub_id: str) -> ModuleType: ...
    def service(self, serv: TS | type[TS]) -> TS: ...

class RootlessPlugin(Plugin):
    @classmethod
    @overload
    def apply(
        cls, id: str, *, default: bool = False
    ) -> Callable[[Callable[[RootlessPlugin], Any]], Callable[[], None]]: ...
    @classmethod
    @overload
    def apply(cls, id: str, func: Callable[[RootlessPlugin], Any], *, default: bool = False) -> Callable[[], None]: ...
    def __init__(self, id: str, func: Callable[[RootlessPlugin], Any], config: dict): ...

class KeepingVariable(Generic[T]):
    obj: T
    _dispose: Callable[[T], None] | None
    def __init__(self, obj: T, dispose: Callable[[T], None] | None = None): ...
    def dispose(self): ...

@overload
def keeping(id_: str, obj: T, *, dispose: Callable[[T], None] | None = None) -> T: ...
@overload
def keeping(id_: str, *, obj_factory: Callable[[], T], dispose: Callable[[T], None] | None = None) -> T: ...



class RegisterNotInPluginError(Exception):
    def __init__(self, func, mod, pid):
        self.msg = f"""\
Handler `{func.__qualname__}` from {func.__module__!r} should define in the same module as the plugin: {mod.__name__!r}.

Please choose one of the following solutions before import it:
 * add {func.__module__!r} to your config file.
 * write the comment after the import statement line: `# entari: plugin`
 * append `load_plugin({func.__module__!r})` before the import statement.
 * call `requires({func.__module__!r})` in the {mod.__name__!r}.
 * write the comment after the import statement line: `# entari: package`
 * call `package({func.__module__!r})` in the {mod.__name__!r} to let it marked as a sub-plugin.\
"""
        try:
            from rich.console import Console
            from rich.panel import Panel

            panel = Panel(
                f"""\
[cyan]Handler [bright_yellow]`{func.__qualname__}`[/] from [blue]{func.__module__!r}\
[/] should define [u bright_white]in the same module[/] as the plugin: [blue]{mod.__name__!r}[/].

Please choose one of the following solutions:
[magenta] *[/] add [bright_green]{func.__module__!r}[/] to your config file.
[magenta] *[/] write the comment after the import statement line: [white]`# entari: plugin`[/]
[magenta] *[/] append [bright_yellow]`load_plugin({func.__module__!r})`[/] [u bright_white]before the import stmt[/].
[magenta] *[/] call [bright_yellow]`requires({func.__module__!r})`[/] in the [blue]{mod.__name__!r}[/].
[magenta] *[/] write the comment after the import statement line: [white]`# entari: package`[/]
[magenta] *[/] call [bright_yellow]`package({func.__module__!r})`[/] in the [blue]{mod.__name__!r}[/].\
""",
                title="Notice",
                style="bold red",
                expand=True,
            )
            csl = Console(force_terminal=True)
            with csl.capture() as capture:
                csl.print(panel)
            self.msg = capture.get()
        except ImportError:
            pass


class StaticPluginDispatchError(Exception):
    pass


class ReusablePluginError(Exception):
    pass



import asyncio
from collections.abc import Awaitable, Callable, Iterable
from typing import Any, Generic, NoReturn, cast, overload
from typing_extensions import TypeVar

from arclet.letoderea import STOP, es, step_out
from satori import ChannelType, Quote
from satori.client.account import Account
from satori.client.protocol import ApiProtocol
from satori.const import Api
from satori.element import At, Element
from satori.model import (
    Channel,
    Guild,
    IterablePageResult,
    Login,
    Member,
    MessageObject,
    MessageReceipt,
    Meta,
    Role,
    Upload,
    User,
)

from .event.base import (
    FriendRequestEvent,
    GuildMemberRequestEvent,
    GuildRequestEvent,
    MessageCreatedEvent,
    MessageEvent,
    Reply,
    SatoriEvent,
)
from .event.send import SendRequest, SendResponse
from .message import MessageChain

TEvent = TypeVar("TEvent", bound=SatoriEvent, default=SatoriEvent)
T = TypeVar("T")


class EntariProtocol(ApiProtocol):
    # fmt: off

    async def send_message(self, channel: str | Channel, message: str | Iterable[str | Element], source: SatoriEvent | None = None, at_sender: At | None = None, reply_to: Quote | None = None) -> list[MessageReceipt]:  # noqa: E501
        """发送消息。返回一个 `MessageReceipt` 对象构成的数组。

        Args:
            channel (str | Channel): 要发送的频道 ID
            message (str | Iterable[str | Element]): 要发送的消息
            source (SatoriEvent | None): 源事件
            at_sender (At | None): 是否 @ 发送者，默认为 None
            reply_to (Quote | None): 是否作为回复发送，默认为 None

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        channel_id = channel.id if isinstance(channel, Channel) else channel
        return await self.message_create(channel_id=channel_id, content=message, source=source, at_sender=at_sender, reply_to=reply_to)  # noqa: E501

    async def send_private_message(self, user: str | User, message: str | Iterable[str | Element], source: SatoriEvent | None = None, at_sender: At | None = None, reply_to: Quote | None = None) -> list[MessageReceipt]:  # noqa: E501
        """发送私聊消息。返回一个 `MessageReceipt` 对象构成的数组。

        Args:
            user (str | User): 要发送的用户 ID
            message (str | Iterable[str | Element]): 要发送的消息
            source (SatoriEvent | None): 源事件
            at_sender (At | None): 是否 @ 发送者，默认为 None
            reply_to (Quote | None): 是否作为回复发送，默认为 None

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        user_id = user.id if isinstance(user, User) else user
        channel = await self.user_channel_create(user_id=user_id)
        return await self.message_create(channel_id=channel.id, content=message, source=source, at_sender=at_sender, reply_to=reply_to)  # noqa: E501

    async def message_create(self, channel_id: str, content: str | Iterable[str | Element], source: SatoriEvent | None = None, at_sender: At | None = None, reply_to: Quote | None = None) -> list[MessageReceipt]:  # noqa: E501
        """发送消息。返回一个 `MessageReceipt` 对象构成的数组。

        Args:
            channel_id (str): 频道 ID
            content (str | Iterable[str | Element]): 消息内容
            source (SatoriEvent | None): 源事件
            at_sender (At | None): 是否 @ 发送者，默认为 None
            reply_to (Quote | None): 是否作为回复发送，默认为 None

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        msg: MessageChain = MessageChain.of(content) if isinstance(content, str) else MessageChain(content)
        if at_sender:
            msg.insert(0, at_sender)
        if reply_to:
            msg.insert(0, reply_to)
        sess = None
        if source:
            sess = Session(self.account.custom(protocol_cls=EntariProtocol), source)
            sess.elements = msg
        res = await es.post(ev := SendRequest(self.account.custom(protocol_cls=EntariProtocol), channel_id, msg, sess))
        msg = sess._content if sess and sess._content else ev.message
        if res:
            if res.value is False:
                return []
            elif isinstance(res.value, MessageChain):
                msg = res.value
        send = str(msg)
        res = await self.call_api(
            Api.MESSAGE_CREATE,
            {"channel_id": channel_id, "content": send},
        )
        res = cast("list[dict]", res)
        resp = [MessageReceipt.parse(i) for i in res]
        await es.publish(SendResponse(self.account.custom(protocol_cls=EntariProtocol), channel_id, msg, resp, sess))
        return resp

    # fmt: on


class Session(Generic[TEvent]):
    """在 Satori-Python 的 Session 的基础上存储了当次事件的信息"""

    def __init__(self, account: Account[EntariProtocol], event: TEvent):
        self.account = account
        self.event = event
        self.type = event.type
        self._content = None
        if isinstance(event, MessageEvent):
            self._content = MessageChain(event.message.message)
        self.reply: Reply | None = None

    @overload
    async def prompt(
        self,
        message: str | Iterable[str | Element] | None = None,
        /,
        *,
        timeout: float = 120,
        timeout_message: str | Iterable[str | Element] = "等待超时",
        block: bool = True,
    ) -> MessageChain | None:
        """等待当前会话的下一次输入并返回

        Args:
            message: 等待前用于提示的消息
            timeout: 等待超时时间
            timeout_message: 超时后发送的消息
            block: 是否阻塞后续的消息传递

        Returns:
            MessageChain | None: 回复的消息，若超时则返回 None
        """
        ...

    @overload
    async def prompt(
        self,
        handler: Callable[..., Awaitable[T]],
        message: str | Iterable[str | Element] | None = None,
        /,
        *,
        timeout: float = 120,
        timeout_message: str | Iterable[str | Element] = "等待超时",
        block: bool = True,
    ) -> T | None:
        """处理当前会话的下一次输入并返回

        Args:
            handler: 用于处理输入的函数
            message: 等待前用于提示的消息
            timeout: 等待超时时间
            timeout_message: 超时后发送的消息
            block: 是否阻塞后续的消息传递

        Returns:
            T | None: 回复的内容，若超时则返回 None
        """
        ...

    async def prompt(
        self,
        *args,
        timeout: float = 120,
        timeout_message: str | Iterable[str | Element] = "等待超时",
        block: bool = True,
    ):
        if not args:
            handler, message = None, None
        elif len(args) == 1:
            if callable(args[0]):
                handler, message = args[0], None
            else:
                handler, message = None, args[0]
        else:
            handler, message = args[0], args[1]
        if message:
            await self.send(message)

        if handler:
            step = step_out(MessageCreatedEvent, handler, block=block)
        else:

            async def waiter(content: MessageChain, session: Session[MessageCreatedEvent]):
                if self.event.user and self.event.user.id == session.event.user.id:
                    if session.event.channel.type == ChannelType.DIRECT:
                        return content
                    if self.event.channel and session.event.channel.id == self.event.channel.id:
                        return content

            waiter.__annotations__ = {"content": MessageChain, "session": self.__class__}

            step = step_out(MessageCreatedEvent, waiter, block=block)

        result = await step.wait(timeout=timeout)
        if result is None:
            await self.send(timeout_message)
            return None
        return result

    def stop(self) -> NoReturn:
        raise STOP

    @property
    def quote(self):
        if isinstance(self.event, MessageEvent):
            return self.event.quote
        return None

    @property
    def user(self) -> User:
        if not self.event.user:
            raise RuntimeError(f"Event {self.event.type!r} has no User")
        return self.event.user

    @property
    def guild(self) -> Guild:
        if not self.event.guild:
            raise RuntimeError(f"Event {self.event.type!r} has no Guild")
        return self.event.guild

    @property
    def channel(self) -> Channel:
        if not self.event.channel:
            raise RuntimeError(f"Event {self.event.type!r} has no Channel")
        return self.event.channel

    @property
    def member(self) -> Member:
        if not self.event.member:
            raise RuntimeError(f"Event {self.event.type!r} has no Member")
        return self.event.member

    @property
    def content(self) -> str:
        if self._content:
            return str(self._content)
        raise RuntimeError(f"Event {self.event.type!r} has no Content")

    @content.setter
    def content(self, value: str):
        self._content = MessageChain(value)

    @property
    def elements(self) -> MessageChain:
        if self._content:
            return self._content
        raise RuntimeError(f"Event {self.event.type!r} has no Content")

    @elements.setter
    def elements(self, value: MessageChain):
        self._content = value

    def _resolve(self, at_sender: bool, reply_to: bool):
        at = reply = None
        if at_sender:
            if not self.event.user:
                raise RuntimeError("Event has no User to @!")
            at = At(self.event.user.id)
        if reply_to:
            if not isinstance(self.event, MessageEvent):
                raise RuntimeError("Event cannot be replied to!")
            if not self.event.message:
                raise RuntimeError("Event cannot be replied to!")
            reply = Quote(self.event.message.id)
        return at, reply

    # fmt: off

    async def send(self, message: str | Iterable[str | Element], at_sender: bool = False, reply_to: bool = False) -> list[MessageReceipt]:  # noqa: E501
        """发送消息。返回一个 `MessageObject` 对象构成的数组。

        Args:
            message: 要发送的消息
            at_sender: 是否 @ 发送者，默认为 False
            reply_to: 是否作为回复发送，默认为 False

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        if not self.event._origin.channel:
            raise RuntimeError("Event has no Channel context!")
        return await self.account.protocol.send_message(self.event._origin.channel.id, message, self.event, *self._resolve(at_sender, reply_to))  # noqa: E501

    async def send_message(self, message: str | Iterable[str | Element], at_sender: bool = False, reply_to: bool = False) -> list[MessageReceipt]:  # noqa: E501
        """发送消息。返回一个 `MessageObject` 对象构成的数组。

        Args:
            message: 要发送的消息
            at_sender: 是否 @ 发送者，默认为 False
            reply_to: 是否作为回复发送，默认为 False

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        if not self.event.channel:
            raise RuntimeError("Event has no Channel context!")
        return await self.account.protocol.send_message(self.event.channel.id, message, self.event, *self._resolve(at_sender, reply_to))  # noqa: E501

    async def send_private_message(self, message: str | Iterable[str | Element], user_id: str | None = None, at_sender: bool = False, reply_to: bool = False) -> list[MessageReceipt]:  # noqa: E501
        """发送私聊消息。返回一个 `MessageObject` 对象构成的数组。

        Args:
            message: 要发送的消息
            user_id: 要发送的用户 ID，默认为 None（即发送给当前事件的用户）
            at_sender: 是否 @ 发送者，默认为 False
            reply_to: 是否作为回复发送，默认为 False

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        channel = await self.user_channel_create(user_id)
        return await self.account.protocol.send_message(channel.id, message, self.event, *self._resolve(at_sender, reply_to))  # noqa: E501

    # fmt: on

    async def update_message(self, message: str | Iterable[str | Element]):
        """更新消息。

        Args:
            message (str | Iterable[str | Element]): 更新后的消息

        Returns:
            None: 该方法无返回值
        """
        if not self.event.channel:
            raise RuntimeError("Event has no Channel context!")
        if not self.event.message:
            raise RuntimeError("Event cannot update message")
        return await self.account.protocol.update_message(self.event.channel, self.event.message.id, message)

    async def message_create(self, content: str) -> list[MessageReceipt]:
        """发送消息。返回一个 `MessageObject` 对象构成的数组。

        Args:
            content (str): 消息内容

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.protocol.message_create(self.event.channel.id, content, self.event)

    async def message_get(self, message_id: str) -> MessageObject:
        """获取特定消息。返回一个 `MessageObject` 对象。

        Args:
            message_id (str): 消息 ID

        Returns:
            MessageObject: `MessageObject` 对象
        """
        if not self.event.channel:
            raise RuntimeError("Event has no Channel context!")
        return await self.account.protocol.message_get(self.event.channel.id, message_id)

    async def message_delete(self, message_id: str, delay: float = -1) -> None:
        """撤回特定消息。

        Args:
            message_id (str): 消息 ID
            delay (float, optional): 延迟撤回时间，单位为秒，默认为 -1（表示不延迟）

        Returns:
            None: 该方法无返回值
        """
        if delay > 0:
            await asyncio.sleep(delay)
        if not self.event.channel:
            raise RuntimeError("Event has no Channel context!")
        await self.account.protocol.message_delete(self.event.channel.id, message_id)

    async def message_update(self, content: str) -> None:
        """编辑特定消息。

        Args:
            content (str): 消息内容

        Returns:
            None: 该方法无返回值
        """
        if not self.event.channel:
            raise RuntimeError("Event has no Channel context!")
        if not self.event.message:
            raise RuntimeError("Event cannot update message")
        await self.account.protocol.message_update(
            self.event.channel.id,
            self.event.message.id,
            content,
        )

    async def channel_create(self, data: Channel) -> Channel:
        """创建群组频道。返回一个 Channel 对象。

        Args:
            data (Channel): 频道数据

        Returns:
            Channel: `Channel` 对象
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to create channel!")
        return await self.account.protocol.channel_create(self.event.guild.id, data)

    async def channel_get(self, channel_id: str) -> Channel:
        """根据 ID 获取频道。返回一个 `Channel` 对象。

        Args:
            channel_id (str): 频道 ID

        Returns:
            Channel: `Channel` 对象
        """
        return await self.account.protocol.channel_get(channel_id)

    def channel_list(self, next_token: str | None = None) -> IterablePageResult[Channel]:
        """获取群组中的全部频道。返回一个 Channel 的分页列表。

        Args:
            next_token (str | None, optional): 分页令牌，默认为空

        Returns:
            IterablePageResult[Channel]: `Channel` 的分页列表
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to list channel!")
        return self.account.protocol.channel_list(self.event.guild.id, next_token)

    async def channel_update(self, data: Channel) -> None:
        """修改群组频道。

        Args:
            data (Channel): 频道数据

        Returns:
            None: 该方法无返回值
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot use to update channel!")
        return await self.account.protocol.channel_update(self.event.channel.id, data)

    async def channel_delete(self) -> None:
        """删除群组频道。

        Returns:
            None: 该方法无返回值
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot use to delete channel!")
        return await self.account.protocol.channel_delete(self.event.channel.id)

    async def user_channel_create(self, user_id: str | None = None) -> Channel:
        """创建一个私聊频道。返回一个 Channel 对象。

        Args:
            user_id (str | None, optional): 用户 ID，默认为 None（即当前事件的用户）

        Returns:
            Channel: `Channel` 对象
        """
        if not user_id and not self.event.user:
            raise RuntimeError("Event cannot use to create user channel!")
        return await self.account.protocol.user_channel_create(
            user_id or self.event.user.id, self.event.guild.id if self.event.guild else None  # type: ignore
        )

    def guild_member_list(self, next_token: str | None = None) -> IterablePageResult[Member]:
        """获取群组成员列表。返回一个 Member 的分页列表。

        Args:
            next_token (str | None, optional): 分页令牌，默认为空

        Returns:
            IterablePageResult[Member]: `Member` 的分页列表
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to list member!")
        return self.account.protocol.guild_member_list(self.event.guild.id, next_token)

    async def guild_member_get(self, user_id: str | None = None) -> Member:
        """获取群成员信息。返回一个 `Member` 对象。

        Args:
            user_id (str): 用户 ID

        Returns:
            Member: `Member` 对象
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to get member!")
        if user_id:
            return await self.account.protocol.guild_member_get(self.event.guild.id, user_id)
        if not self.event.user:
            raise RuntimeError("Event cannot use to get member!")
        return await self.account.protocol.guild_member_get(self.event.guild.id, self.event.user.id)

    async def guild_member_kick(self, user_id: str | None = None, permanent: bool = False) -> None:
        """将某个用户踢出群组。

        Args:
            user_id (str): 用户 ID
            permanent (bool, optional): 是否永久踢出 (无法再次加入群组)，默认为 False

        Returns:
            None: 该方法无返回值
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to kick member!")
        if user_id:
            return await self.account.protocol.guild_member_kick(self.event.guild.id, user_id, permanent)
        if not self.event.user:
            raise RuntimeError("Event cannot use to kick member!")
        return await self.account.protocol.guild_member_kick(self.event.guild.id, self.event.user.id, permanent)

    async def guild_member_role_set(self, role_id: str, user_id: str | None = None) -> None:
        """设置群组内用户的角色。

        Args:
            user_id (str): 用户 ID
            role_id (str): 角色 ID

        Returns:
            None: 该方法无返回值
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to guild member role set!")
        if user_id:
            return await self.account.protocol.guild_member_role_set(self.event.guild.id, user_id, role_id)
        if not self.event.user:
            raise RuntimeError("Event cannot use to guild member role set!")
        return await self.account.protocol.guild_member_role_set(self.event.guild.id, self.event.user.id, role_id)

    async def guild_member_role_unset(self, role_id: str, user_id: str | None = None) -> None:
        """取消群组内用户的角色。

        Args:
            user_id (str): 用户 ID
            role_id (str): 角色 ID

        Returns:
            None: 该方法无返回值
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to guild member role unset!")
        if user_id:
            return await self.account.protocol.guild_member_role_unset(self.event.guild.id, user_id, role_id)
        if not self.event.user:
            raise RuntimeError("Event cannot use to guild member role unset!")
        return await self.account.protocol.guild_member_role_unset(self.event.guild.id, self.event.user.id, role_id)

    def guild_role_list(self, next_token: str | None = None) -> IterablePageResult[Role]:
        """获取群组角色列表。返回一个 Role 的分页列表。

        Args:
            next_token (str | None, optional): 分页令牌，默认为空

        Returns:
            IterablePageResult[Role]: `Role` 的分页列表
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to list role!")
        return self.account.protocol.guild_role_list(self.event.guild.id, next_token)

    async def guild_role_create(self, role: Role) -> Role:
        """创建群组角色。返回一个 Role 对象。

        Args:
            role (Role): 角色数据

        Returns:
            Role: `Role` 对象
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to create role!")
        return await self.account.protocol.guild_role_create(self.event.guild.id, role)

    async def guild_role_update(self, role_id: str, role: Role) -> None:
        """修改群组角色。

        Args:
            role_id (str): 角色 ID
            role (Role): 角色数据

        Returns:
            None: 该方法无返回值
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to update role!")
        return await self.account.protocol.guild_role_update(self.event.guild.id, role_id, role)

    async def guild_role_delete(self, role_id: str) -> None:
        """删除群组角色。

        Args:
            role_id (str): 角色 ID

        Returns:
            None: 该方法无返回值
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to delete role!")
        return await self.account.protocol.guild_role_delete(self.event.guild.id, role_id)

    async def reaction_create(self, emoji: str, message_id: str | None = None) -> None:
        """向特定消息添加表态。

        Args:
            message_id (str | None, optional): 消息 ID
            emoji (str): 表态名称

        Returns:
            None: 该方法无返回值
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not message_id and not self.event.message:
            raise RuntimeError("Event cannot create reaction")
        return await self.account.protocol.reaction_create(self.event.channel.id, message_id or self.event.message.id, emoji)  # type: ignore  # noqa: E501

    async def reaction_delete(self, emoji: str, message_id: str | None = None, user_id: str | None = None) -> None:
        """从特定消息删除某个用户添加的特定表态。

        如果没有传入用户 ID 则表示删除自己的表态。

        Args:
            message_id (str | None, optional): 消息 ID
            emoji (str): 表态名称
            user_id (str | None, optional): 用户 ID，默认为 None

        Returns:
            None: 该方法无返回值
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not message_id and not self.event.message:
            raise RuntimeError("Event cannot delete reaction")
        return await self.account.protocol.reaction_delete(self.event.channel.id, message_id or self.event.message.id, emoji, user_id)  # type: ignore  # noqa: E501

    async def reaction_clear(self, emoji: str | None = None, message_id: str | None = None) -> None:
        """从特定消息清除某个特定表态。

        如果没有传入表态名称则表示清除所有表态。

        Args:
            message_id (str | None, optional): 消息 ID
            emoji (str | None, optional): 表态名称，默认为 None

        Returns:
            None: 该方法无返回值
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not message_id and not self.event.message:
            raise RuntimeError("Event cannot clear reaction")
        return await self.account.protocol.reaction_clear(
            self.event.channel.id,
            message_id or self.event.message.id,  # type: ignore
            emoji,
        )

    def reaction_list(
        self, emoji: str, message_id: str | None = None, next_token: str | None = None
    ) -> IterablePageResult[User]:
        """获取添加特定消息的特定表态的用户列表。返回一个 User 的分页列表。

        Args:
            message_id (str | None, optional): 消息 ID
            emoji (str): 表态名称
            next_token (str | None, optional): 分页令牌，默认为空

        Returns:
            IterablePageResult[User]: `User` 的分页列表
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not message_id and not self.event.message:
            raise RuntimeError("Event cannot list reaction")
        return self.account.protocol.reaction_list(
            self.event.channel.id, message_id or self.event.message.id, emoji, next_token  # type: ignore
        )

    async def request_approve(self, approve: bool, comment: str = ""):
        """处理请求。

        Args:
            approve (bool): 是否通过请求
            comment (str): 备注信息
        Returns:
            None: 该方法无返回值
        """
        if isinstance(self.event, FriendRequestEvent):
            return await self.friend_approve(approve, comment)
        if isinstance(self.event, GuildRequestEvent):
            return await self.guild_approve(approve, comment)
        if isinstance(self.event, GuildMemberRequestEvent):
            return await self.guild_member_approve(approve, comment)
        raise RuntimeError("Event cannot approve request")

    async def friend_approve(self, approve: bool, comment: str = ""):
        """处理好友申请。

        Args:
            approve (bool): 是否通过请求
            comment (str): 备注信息

        Returns:
            None: 该方法无返回值
        """
        if not isinstance(self.event, FriendRequestEvent):
            raise RuntimeError("Event cannot approve friend request")
        return await self.account.protocol.friend_approve(self.event.message.id, approve, comment)

    async def guild_approve(self, approve: bool, comment: str = ""):
        """处理来自群组的邀请。

        Args:
            approve (bool): 是否通过请求
            comment (str): 备注信息

        Returns:
            None: 该方法无返回值
        """
        if not isinstance(self.event, GuildRequestEvent):
            raise RuntimeError("Event cannot approve guild request")
        return await self.account.protocol.guild_approve(self.event.message.id, approve, comment)

    async def guild_member_approve(self, approve: bool, comment: str = ""):
        """处理来自群组的加群请求。

        Args:
            approve (bool): 是否通过请求
            comment (str): 备注信息

        Returns:
            None: 该方法无返回值
        """
        if not isinstance(self.event, GuildMemberRequestEvent):
            raise RuntimeError("Event cannot approve guild member request")
        return await self.account.protocol.guild_member_approve(self.event.message.id, approve, comment)

    async def channel_mute(self, duration: float = 60) -> None:
        """禁言群组频道。

        如果传入的禁言时长为 0 则表示解除禁言。

        Args:
            duration (float, optional): 禁言时长 (秒)，默认为 60 秒
        Returns:
            None: 该方法无返回值
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot use to mute channel!")
        return await self.account.protocol.channel_mute(self.event.channel.id, duration)

    async def guild_get(self, guild_id: str) -> Guild:
        """根据 ID 获取群组。返回一个 `Guild` 对象。

        Args:
            guild_id (str): 群组 ID

        Returns:
            Guild: `Guild` 对象
        """
        return await self.account.protocol.guild_get(guild_id)

    def guild_list(self, next_token: str | None = None) -> IterablePageResult[Guild]:
        """获取当前用户加入的全部群组。返回一个 Guild 的分页列表。

        Args:
            next_token (str | None, optional): 分页令牌，默认为空

        Returns:
            IterablePageResult[Guild]: `Guild` 的分页列表
        """
        return self.account.protocol.guild_list(next_token)

    async def guild_member_mute(self, user_id: str, duration: float = 60) -> None:
        """禁言群组成员。

        如果传入的禁言时长为 0 则表示解除禁言。

        Args:
            user_id (str): 用户 ID
            duration (float, optional): 禁言时长 (秒)，默认为 60 秒

        Returns:
            None: 该方法无返回值
        """
        if not self.event.guild:
            raise RuntimeError("Event cannot use to mute guild member!")
        return await self.account.protocol.guild_member_mute(self.event.guild.id, user_id, duration)

    async def login_get(self) -> Login:
        """获取当前登录信息。返回一个 `Login` 对象。

        Returns:
            Login: `Login` 对象
        """
        return await self.account.protocol.login_get()

    async def user_get(self, user_id: str) -> User:
        """获取用户信息。返回一个 `User` 对象。

        Args:
            user_id (str): 用户 ID

        Returns:
            User: `User` 对象
        """
        return await self.account.protocol.user_get(user_id)

    def friend_list(self, next_token: str | None = None) -> IterablePageResult[User]:
        """获取好友列表。返回一个 User 的分页列表。

        Args:
            next_token (str | None, optional): 分页令牌，默认为空

        Returns:
            IterablePageResult[User]: `User` 的分页列表
        """
        return self.account.protocol.friend_list(next_token)

    async def internal(self, action: str, method: str = "POST", **kwargs) -> Any:
        """内部接口调用。

        Args:
            action (str): 内部接口名称
            method (str, optional): 请求方法，默认为 POST
            **kwargs: 参数
        """
        return await self.account.protocol.internal(action, method, **kwargs)

    async def meta_get(self) -> Meta:
        """获取元信息。返回一个 `Meta` 对象。

        Returns:
            Meta: `Meta` 对象
        """
        return await self.account.protocol.meta_get()

    @overload
    async def upload_create(self, *uploads: Upload) -> list[str]: ...
    @overload
    async def upload_create(self, **uploads: Upload) -> dict[str, str]: ...

    async def upload_create(self, *args: Upload, **kwargs: Upload):  # type: ignore
        """上传文件。

        如果要发送的消息中含有图片或其他媒体资源，\
            可以使用此 API 将文件上传至 Satori 服务器并转换为 URL，以便在消息编码中使用。
        """
        return await self.account.protocol.upload_create(*args, **kwargs)

    upload = upload_create

    async def download(self, url: str):
        """访问内部链接。"""
        return await self.account.protocol.download(url)

    async def request_internal(self, url: str, method: str = "GET", **kwargs) -> dict:
        """访问内部链接。"""
        return await self.account.protocol.request_internal(url, method, **kwargs)



from abc import ABCMeta, abstractmethod
from dataclasses import asdict, fields
from typing import Any, Generic, TypeVar

_config_model_actions: dict[type, type["ConfigModelAction"]] = {}

C = TypeVar("C")


class ConfigModelAction(Generic[C], metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def load(cls, data: dict[str, Any], t: type[C]) -> C:
        """
        Validate the configuration data and return a model instance.
        """
        pass

    @classmethod
    @abstractmethod
    def dump(cls, obj: C) -> dict[str, Any]:
        """
        Update the configuration data from the model instance.
        """
        pass

    @classmethod
    @abstractmethod
    def keys(cls, obj: C) -> list[str]:
        """
        Get the keys of the configuration model.
        """
        pass

    @classmethod
    @abstractmethod
    def schema(cls, t: type[C], ref_root: str = "/") -> dict[str, Any]:
        """
        Get the schema of the configuration model.
        """
        pass

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        if cls.__orig_bases__[0].__args__[0] is C:  # type: ignore
            raise TypeError("Subclass of ConfigModelAction must be generic.")
        base = cls.__orig_bases__[0].__args__[0]  # type: ignore
        _config_model_actions[base] = cls
        return cls


class Proxy:
    def __init__(self, origin, updater):
        self.__origin = origin
        self.__updater = updater

    def __getattr__(self, item):
        res = getattr(self.__origin, item)
        if isinstance(res, (list, tuple, set, dict, self.__origin.__class__)):  # noqa: UP038
            return Proxy(res, self.__updater)
        return res

    def __getitem__(self, item):
        res = self.__origin[item]
        if isinstance(res, (list, tuple, set, dict, self.__origin.__class__)):  # noqa: UP038
            return Proxy(res, self.__updater)
        return res

    def __setattr__(self, key, value):
        if key in {"_Proxy__origin", "_Proxy__updater"}:
            super().__setattr__(key, value)
        else:
            setattr(self.__origin, key, value)
            self.__updater()

    def __setitem__(self, key, value):
        self.__origin[key] = value
        self.__updater()

    def __repr__(self):
        return self.__origin.__repr__()

    def __str__(self):
        return self.__origin.__str__()

    def __len__(self):
        return len(self.__origin)

    def __contains__(self, item):
        return item in self.__origin

    def __iter__(self):
        return self.__origin.__iter__()

    def __eq__(self, other):
        if isinstance(other, Proxy):
            return self.__origin == other._Proxy__origin
        return self.__origin == other

    def __bool__(self):
        return bool(self.__origin)


def config_model_validate(base: type[C], data: dict[str, Any]) -> C:
    data = {k: v for k, v in data.items() if not k.startswith("$")}
    for b in base.__mro__[-2::-1]:
        if b in _config_model_actions:
            return _config_model_actions[b].load(data, base)
    return base(**data)  # type: ignore


def config_model_dump(obj: Any) -> dict[str, Any]:
    for b in obj.__class__.__mro__[-2::-1]:
        if b in _config_model_actions:
            return _config_model_actions[b].dump(obj)
    return asdict(obj)  # type: ignore


def config_model_keys(obj: Any) -> list[str]:
    for b in obj.__class__.__mro__[-2::-1]:
        if b in _config_model_actions:
            return _config_model_actions[b].keys(obj)
    return [field_.name for field_ in fields(obj)]  # type: ignore


def config_model_schema(base: type[C], ref_root: str = "/") -> dict[str, Any]:
    for b in base.__mro__[-2::-1]:
        if b in _config_model_actions:
            return _config_model_actions[b].schema(base, ref_root)
    return {field_.name: field_.type for field_ in fields(base)}  # type: ignore



from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator, Awaitable, Callable, Generator
from typing import TypeAlias, TypeVar, cast, overload
from weakref import WeakValueDictionary

from arclet.alconna import Alconna, Arg, Args, CommandMeta, Namespace, command_manager, config
from arclet.alconna.tools.construct import AlconnaString, alconna_from_format
from arclet.alconna.typing import TAValue
import arclet.letoderea as le
from arclet.letoderea import RESULT, ExitState, Scope, Subscriber, make_event
from arclet.letoderea.provider import TProviders, get_providers
from arclet.letoderea.typing import Contexts, Result
from nepattern import DirectPattern
from satori.element import Text
from tarina.trie import CharTrie

from ..config import BasicConfModel, config_model_validate, model_field
from ..event.base import MessageCreatedEvent
from ..event.command import CommandExecute
from ..event.config import ConfigReload
from ..message import MessageChain
from ..plugin import RootlessPlugin, get_plugin, metadata, plugin_config
from ..session import Session
from .argv import MessageArgv  # noqa: F401
from .model import CommandResult, Match, Query
from .plugin import mount
from .provider import AlconnaProviderFactory, AlconnaSuppiler, MessageJudges

_BaseM: TypeAlias = str | MessageChain | None
_M: TypeAlias = _BaseM | Generator[_BaseM, None, None] | AsyncGenerator[_BaseM, None] | Awaitable[_BaseM]
_RM: TypeAlias = _BaseM | AsyncGenerator[_BaseM, None]
TM = TypeVar("TM", bound=_M)


def get_cmd(target: Subscriber):
    if sup := target.get_propagator(AlconnaSuppiler):
        return sup.cmd
    raise ValueError("Subscriber has no command.")


@make_event(name="entari.event/internal/command_dispatch")
class CommandDispatch:
    providers = [*get_providers(MessageCreatedEvent), *get_providers(CommandExecute), AlconnaProviderFactory()]


async def _after_execute(ctx: Contexts, session: Session | None = None):
    result: _RM = ctx[RESULT]
    if result is not None:
        if isinstance(result, AsyncGenerator):
            msg = None
            async for msg in result:
                if session and msg is not None:
                    await session.send(msg)
            return Result(msg)
        if session:
            await session.send(result)
        return Result(result)


class EntariCommands:
    __namespace__ = "Entari"

    def __init__(
        self,
        block: bool = True,
        need_notice_me: bool = False,
        need_reply_me: bool = False,
        use_config_prefix: bool = True,
    ):
        self.trie: CharTrie[str] = CharTrie()
        self.scope = Scope("entari.command")
        self.block = block
        self.judge = MessageJudges(need_notice_me, need_reply_me, use_config_prefix)
        config.namespaces["Entari"] = Namespace(
            self.__namespace__,
            to_text=lambda x: x.text if x.__class__ is Text else None,
            converter=lambda x: MessageChain(x),
        )
        self.subscribers = WeakValueDictionary[str, Subscriber]()

        le.on(CommandExecute, self.execute)

    @property
    def all_helps(self) -> str:
        return command_manager.all_command_help(namespace=self.__namespace__)

    def get_help(self, command: str) -> str:
        return command_manager.get_command(f"{self.__namespace__}::{command}").get_help()

    async def execute(self, message: MessageChain, ctx: Contexts):
        msg = str(message).lstrip()
        if not msg:
            return
        subs = self.subscribers
        if matches := list(self.trie.prefixes(msg)):
            subs = {res.value: self.subscribers[res.value] for res in matches if res.value in self.subscribers}
        results = await asyncio.gather(*(sub.handle(ctx.copy(), inner=True) for sub in subs.values()))
        for result in results:
            if result is ExitState.stop:
                continue
            if result is ExitState.block:
                return
            if result is not None:
                if self.block:
                    return ExitState.block.finish(cast(str | MessageChain, result))
                return cast(str | MessageChain, result)

    def command(self, cmd: str, help_text: str | None = None, providers: TProviders | None = None):
        class Command(AlconnaString):
            def __call__(_cmd_self, func: Callable[..., TM]) -> Subscriber[TM]:
                return self.on(_cmd_self.build(), providers)(func)

        return Command(cmd, help_text)

    # fmt: off
    @overload
    def on(self, cmd: Alconna, providers: TProviders | None = None) -> Callable[[Callable[..., TM]], Subscriber[TM]]: ...  # noqa: E501

    @overload
    def on(self, cmd: str, providers: TProviders | None = None, *, args: dict[str, TAValue | Args | Arg] | None = None, meta: CommandMeta | None = None) -> Callable[[Callable[..., TM]], Subscriber[TM]]: ...  # noqa: E501

    def on(self, cmd: Alconna | str, providers: TProviders | None = None, *, args: dict[str, TAValue | Args | Arg] | None = None, meta: CommandMeta | None = None) -> Callable[[Callable[..., TM]], Subscriber[TM]]:  # noqa: E501
        # fmt: on
        plg = get_plugin(1, optional=True)
        providers = providers or []

        def wrapper(func: Callable[..., TM]) -> Subscriber[TM]:
            nonlocal meta
            if isinstance(cmd, str):
                if not meta and func.__doc__:
                    meta = CommandMeta(func.__doc__)
                mapping = {arg.name: arg.value for arg in Args.from_callable(func)[0]}
                mapping.update(args or {})  # type: ignore
                _command = alconna_from_format(cmd, mapping, meta, union=False)
                _command.reset_namespace(self.__namespace__)
                key = _command.name + "".join(
                    f" {arg.value.target}" for arg in _command.args if isinstance(arg.value, DirectPattern)
                )
                if plg:
                    target = plg.dispatch(CommandDispatch).handle(func, providers=providers)
                    plg._extra.setdefault("commands", []).append(([], _command.command))
                else:
                    target = self.scope.register(func, CommandDispatch, providers=providers)
                target.propagate(AlconnaSuppiler(_command))
                target.propagate(_after_execute, priority=0)
                self.trie[key] = target.id
                self.subscribers[target.id] = target

                def _remove(_):
                    command_manager.delete(get_cmd(_))
                    self.trie.pop(key, None)  # type: ignore
                    self.subscribers.pop(target.id, None)

                target._attach_disposes(_remove)
                return target

            _command = cast(Alconna, cmd)
            if not isinstance(cmd.command, str):
                raise TypeError("Command name must be a string.")
            _command.reset_namespace(self.__namespace__)
            keys = []
            if not _command.prefixes:
                keys.append(_command.command)
            elif not all(isinstance(i, str) for i in _command.prefixes):
                raise TypeError("Command prefixes must be a list of string.")
            else:
                for prefix in cast(list[str], _command.prefixes):
                    keys.append(prefix + _command.command)

            if plg:
                target = plg.dispatch(CommandDispatch).handle(func, providers=providers)
                plg._extra.setdefault("commands", []).append((_command.prefixes, _command.command))
            else:
                target = self.scope.register(func, providers=providers)
            target.propagate(AlconnaSuppiler(_command))
            target.propagate(_after_execute, priority=0)
            self.subscribers[target.id] = target
            for _key in keys:
                self.trie[_key] = target.id

            def _remove(_):
                command_manager.delete(get_cmd(_))
                self.subscribers.pop(target.id, None)
                for _key in keys:
                    self.trie.pop(_key, None)  # type: ignore

            target._attach_disposes(_remove)
            return target
        return wrapper


_commands = EntariCommands()


def config_commands(
    block: bool = True, need_notice_me: bool = False, need_reply_me: bool = False, use_config_prefix: bool = True
):
    _commands.block = block
    _commands.judge.need_notice_me = need_notice_me
    _commands.judge.need_reply_me = need_reply_me
    _commands.judge.use_config_prefix = use_config_prefix


command = _commands.command
on = _commands.on


async def execute(message: str | MessageChain):
    res = await le.post(CommandExecute(message))
    return res.value if res else None


class CommandsConfig(BasicConfModel):
    block: bool = model_field(default=True, description="是否阻断消息传播")
    need_notice_me: bool = model_field(default=False, description="是否需要通知我")
    need_reply_me: bool = model_field(default=False, description="是否需要回复我")
    use_config_prefix: bool = model_field(default=True, description="是否使用配置前缀")


@RootlessPlugin.apply("commands", default=True)
def _(plg: RootlessPlugin):
    metadata(
        "Commands Plugin",
        ["RF-Tar-Railt <rf_tar_railt@qq.com>"],
        config=CommandsConfig,
    )

    conf = plugin_config(CommandsConfig)
    _commands.block = conf.block
    _commands.judge.need_notice_me = conf.need_notice_me
    _commands.judge.need_reply_me = conf.need_reply_me
    _commands.judge.use_config_prefix = conf.use_config_prefix

    plg.dispatch(MessageCreatedEvent).handle(_commands.execute).propagate(_commands.judge)

    @plg.dispatch(ConfigReload).handle
    def update(event: ConfigReload):
        if event.scope != "plugin":
            return
        if event.key != ".commands":
            return
        new_conf = config_model_validate(CommandsConfig, event.value)
        _commands.block = new_conf.block
        _commands.judge.need_notice_me = new_conf.need_notice_me
        _commands.judge.need_reply_me = new_conf.need_reply_me
        _commands.judge.use_config_prefix = new_conf.use_config_prefix
        return True


__all__ = ["_commands", "config_commands", "Match", "Query", "execute", "CommandResult", "mount", "command", "on"]



import asyncio
from collections.abc import Awaitable, Callable
from datetime import datetime
import inspect
from typing import Final, TypeAlias
from typing_extensions import ParamSpec

from arclet.letoderea import STOP, Propagator, enter_if
from tarina import is_coroutinefunction

from . import common, message
from ..message import MessageChain
from ..session import Session
from .common import parse as parse

_SessionFilter: TypeAlias = Callable[[Session], bool] | Callable[[Session], Awaitable[bool]]


P = ParamSpec("P")


def _check_wrapper(func: Callable[P, _SessionFilter]):
    def wrapper(*args: P.args, **kwargs: P.kwargs):
        return enter_if(func(*args, **kwargs))

    return wrapper


class _Filter:
    user = staticmethod(_check_wrapper(common.user))
    guild = staticmethod(_check_wrapper(common.guild))
    channel = staticmethod(_check_wrapper(common.channel))
    self_ = staticmethod(_check_wrapper(common.account))
    platform = staticmethod(_check_wrapper(common.platform))
    direct = enter_if(message.direct_message)
    private = enter_if(message.direct_message)
    direct_message = enter_if(message.direct_message)
    public = enter_if(message.public_message)
    public_message = enter_if(message.public_message)
    notice_me = enter_if(message.notice_me)
    reply_me = enter_if(message.reply_me)
    to_me = enter_if(message.to_me)

    def __call__(self, func: _SessionFilter):
        sig = inspect.signature(func)
        name = next(iter(sig.parameters.values())).name

        if not is_coroutinefunction(func):

            async def _(*args, _func=func, **kwargs):
                return _func(*args, **kwargs)

            func = _
        func.__signature__ = sig.replace(
            parameters=[
                param.replace(annotation=Session) if param.name == name else param for param in sig.parameters.values()
            ]
        )
        return enter_if(func)


filter_: Final[_Filter] = _Filter()
F = filter_


class Interval(Propagator):
    def __init__(self, interval: float, limit_prompt: str | MessageChain | None = None, priority: int = 80):
        self.success = True
        self.last_time = None
        self.interval = interval
        self.priority = priority
        self.limit_prompt = limit_prompt

    async def before(self, session: Session | None = None):
        if not self.last_time:
            return
        self.success = (datetime.now() - self.last_time).total_seconds() > self.interval
        if not self.success:
            if session and self.limit_prompt:
                await session.send(self.limit_prompt)
            return STOP

    async def after(self):
        self.last_time = datetime.now()

    def compose(self):
        yield self.before, True, self.priority
        yield self.after, False, self.priority


class Semaphore(Propagator):
    def __init__(self, count: int, limit_prompt: str | MessageChain | None = None, priority: int = 80):
        self.count = count
        self.limit_prompt = limit_prompt
        self.semaphore = asyncio.Semaphore(count)
        self.priority = priority

    async def before(self, session: Session | None = None):
        if not await self.semaphore.acquire():
            if session and self.limit_prompt:
                await session.send(self.limit_prompt)
            return STOP

    async def after(self):
        self.semaphore.release()

    def compose(self):
        yield self.before, True, self.priority
        yield self.after, False, self.priority



from arclet.letoderea import make_event


@make_event(name="entari.event/plugin/loaded_success")
class PluginLoadedSuccess:
    plugin_id: str


@make_event(name="entari.event/plugin/loaded_failed")
class PluginLoadedFailed:
    plugin_id: str
    error: Exception | None = None
    """若没有异常信息，说明该插件加载失败的原因是插件不存在。"""


@make_event(name="entari.event/plugin/unloaded")
class PluginUnloaded:
    plugin_id: str



import asyncio
from collections.abc import Iterable, Sequence
from dataclasses import asdict
import os
from pathlib import Path
import signal
import sys
from typing import get_args

from arclet.alconna import config as alconna_config
import arclet.letoderea as le
from arclet.letoderea import Contexts, Param, Provider, ProviderFactory, global_providers
from arclet.letoderea.scope import configure
from creart import it
from graia.amnesia.builtins.memcache import MemcacheService
from launart import Launart, Service
from satori import Channel, LoginStatus
from satori.client import App
from satori.client.account import Account
from satori.client.config import Config, WebhookInfo, WebsocketsInfo
from satori.client.protocol import ApiProtocol
from satori.model import Event
from tarina.generic import generic_isinstance, get_origin, is_optional

from .config import BasicConfModel, EntariConfig
from .config.file import LogInfo
from .config.model import config_model_validate
from .event.base import MessageCreatedEvent, event_parse
from .event.config import ConfigReload
from .event.lifespan import AccountUpdate
from .event.send import SendResponse
from .localdata import local_data
from .logger import apply_log_save, enable_rich_except, log
from .plugin import get_plugins, load_plugin, plugin_config, requires
from .plugin.model import PluginMetadata, RootlessPlugin
from .plugin.service import plugin_service
from .session import EntariProtocol, Session


class ApiProtocolProvider(Provider[ApiProtocol]):
    async def __call__(self, context: Contexts):
        if account := context.get("$account"):
            return account.protocol


class SessionProviderFactory(ProviderFactory):
    class _SessionProvider(Provider[Session]):
        def __init__(self, target_type: type | None):
            super().__init__()
            self.target_type = target_type

        async def __call__(self, context: Contexts):
            if self.target_type and not generic_isinstance(context["$event"], self.target_type):
                return
            if "$session" in context and isinstance(context["$session"], Session):
                return context["$session"]
            if "$origin_event" in context and "$account" in context:
                session = Session(context["$account"], context["$event"])
                if "$message_content" in context:
                    session.elements = context["$message_content"]
                if "$message_reply" in context:
                    session.reply = context["$message_reply"]
                context["$session"] = session
                return session

    def validate(self, param: Param):
        if get_origin(param.annotation) is Session:
            args = get_args(param.annotation)
            return self._SessionProvider(args[0] if args else None)
        if is_optional(param.annotation, Session):
            args = get_args(get_args(param.annotation)[0])
            return self._SessionProvider(args[0] if args else None)


class AccountProvider(Provider[Account]):
    async def __call__(self, context: Contexts):
        if "$account" in context:
            return context["$account"]


global_providers.extend([ApiProtocolProvider(), SessionProviderFactory(), AccountProvider()])


class RecordConfig(BasicConfModel):
    to_me_only: bool = False
    """是否仅记录提及 Bot 自己的消息"""
    record_send: bool = False
    """是否记录发送的消息"""


@RootlessPlugin.apply("record_message", default=True)
def record(plg: RootlessPlugin):
    plg.metadata = PluginMetadata(
        "记录消息",
        [{"name": "RF-Tar-Railt", "email": "rf_tar_railt@qq.com"}],
        description="将收到[和发出]的消息打印在日志中",
        config=RecordConfig,
    )
    cfg = plugin_config(RecordConfig)

    # fmt: off

    @plg.dispatch(MessageCreatedEvent).on(priority=0)
    async def log_msg(event: MessageCreatedEvent, is_notice_me: bool = False, is_reply_me: bool = False):
        if cfg.to_me_only and not is_notice_me and not is_reply_me:
            return
        if "guild.plain" in event.login.features or not event.guild or (event.guild and event.guild.id == event.channel.id):  # noqa: E501
            scene = f"[{event.channel.name or event.channel.id}" + (f"({event.channel.id})]" if event.channel.name else "]")  # noqa: E501
        else:
            scene = f"[{event.guild.name or event.guild.id}" + (f"({event.guild.id})" if event.guild.name else "") + f" / {event.channel.name or event.channel.id}" + (f"({event.channel.id})]" if event.channel.name else "]")  # noqa: E501

        log.message.info(f"{scene} {event.member.nick if event.member and event.member.nick else (event.user.name or event.user.id)}({event.user.id}) -> {event.message.content!r}")  # noqa: E501

    if cfg.record_send:

        @plg.dispatch(SendResponse)
        async def log_send(event: SendResponse):
            if event.session:
                ev = event.session.event
                if not ev.channel:
                    ev.channel = Channel(event.channel)
                if "guild.plain" in ev.login.features or not ev.guild or (ev.guild and ev.guild.id == ev.channel.id):  # noqa: E501
                    scene = f"[{ev.channel.name or ev.channel.id}" + (f"({ev.channel.id})]" if ev.channel.name else "]")  # noqa: E501
                else:
                    scene = f"[{ev.guild.name or ev.guild.id}" + (f"({ev.guild.id})" if ev.guild.name else "") + f" / {ev.channel.name or ev.channel.id}" + (f"({ev.channel.id})]" if ev.channel.name else "]")  # noqa: E501
                log.message.info(f"{scene} <- {event.message.display()!r}")
            else:
                log.message.info(f"[{event.channel}] <- {event.message.display()!r}")

    # fmt: on


class Entari(App):
    id = "entari.service"

    @property
    def main_plugin(self):
        return RootlessPlugin.apply("main", default=True)

    @classmethod
    def load(cls, path: str | os.PathLike[str] | None = None):
        return cls.from_config(EntariConfig.load(path))

    @classmethod
    def from_config(cls, config: EntariConfig | None = None):
        if not config:
            config = EntariConfig.instance
        ignore_self_message = config.basic.ignore_self_message
        log_level = config.basic.log.level
        skip_req_missing = config.basic.skip_req_missing
        external_dirs = config.basic.external_dirs
        configs = []
        for conf in config.basic.network:
            if conf.type in ("websocket", "websockets", "ws"):
                configs.append(WebsocketsInfo(**{k: v for k, v in asdict(conf).items() if k != "type"}))
            elif conf.type in ("webhook", "wh", "http"):
                configs.append(WebhookInfo(**{k: v for k, v in asdict(conf).items() if k != "type"}))
        return cls(
            *configs,
            log_level=log_level,
            ignore_self_message=ignore_self_message,
            skip_req_missing=skip_req_missing,
            external_dirs=external_dirs,
            rich_error=config.basic.log.rich_error,
            gen_schema=config.basic.schema,
        )

    def __init__(
        self,
        *configs: Config,
        log_level: str | int = "INFO",
        ignore_self_message: bool = True,
        skip_req_missing: bool = False,
        external_dirs: Sequence[str | os.PathLike[str]] | None = None,
        rich_error: bool = False,
        gen_schema: bool = False,
    ):
        from . import __version__

        configure(skip_req_missing=skip_req_missing)
        super().__init__(*configs, default_api_cls=EntariProtocol)
        if not hasattr(EntariConfig, "instance"):
            EntariConfig.load()
        alconna_config.command_max_count = EntariConfig.instance.basic.cmd_count
        log.set_level(log_level)
        log.short_level = EntariConfig.instance.basic.log.short_level
        if rich_error:
            enable_rich_except()
        log.core.info(f"Entari <b><c>version {__version__}</c></b>")
        self._log_save_dispose = lambda: None
        if EntariConfig.instance.basic.log.save:
            if EntariConfig.instance.basic.log.save is True:
                self._log_save_dispose = apply_log_save(local_data._get_base_log_dir())
            else:
                self._log_save_dispose = apply_log_save(
                    log_dir=local_data._get_base_log_dir(),
                    rotation=EntariConfig.instance.basic.log.save.rotation,
                    compression=EntariConfig.instance.basic.log.save.compression,
                    colorize=EntariConfig.instance.basic.log.save.colorize,
                )
        log.ignores.update(EntariConfig.instance.basic.log.ignores)
        log.core.debug(f"Log level set to <y><c>{log_level}</c></y>")
        log.core.debug(f"Config loaded from <m>{EntariConfig.instance.path}</m>: <w>{EntariConfig.instance.data}</w>")
        self.ignore_self_message = ignore_self_message
        self.register(self.handle_event)
        self.lifecycle(self.account_hook)
        self._ref_tasks = set()
        self.gen_schema = gen_schema

        le.on(ConfigReload, self.reset_self)

        self._path_scale = ()
        _external = [str(Path(d).resolve()) for d in external_dirs or []]
        if _external:
            sys.path.extend(_external)
            self._path_scale = (len(sys.path) - len(_external), len(sys.path))
            log.core.debug(f"Added external dirs: {_external}")

    @property
    def config(self):
        return EntariConfig.instance

    def reset_self(self, scope, key, value):
        if scope != "basic":
            return
        if key == "log_level":
            log.set_level(value)
            log.core.debug(f"Log level set to <y><c>{value}</c></y>")
        elif key == "log":
            new_conf = config_model_validate(LogInfo, value)
            log.set_level(new_conf.level)
            log.core.debug(f"Log level set to <y><c>{new_conf.level}</c></y>")
            log.ignores.clear()
            log.ignores.update(new_conf.ignores)
            self._log_save_dispose()
            if new_conf.save:
                if new_conf.save is True:
                    self._log_save_dispose = apply_log_save(local_data._get_base_log_dir())
                else:
                    self._log_save_dispose = apply_log_save(
                        local_data._get_base_log_dir(),
                        new_conf.save.rotation,
                        new_conf.save.compression,
                        new_conf.save.colorize,
                    )
        elif key == "ignore_self_message":
            self.ignore_self_message = value
        elif key == "network":
            for conn in self.connections:
                it(Launart).remove_component(conn)
            self.connections.clear()
            for conf in value:
                if conf["type"] in ("websocket", "websockets", "ws"):
                    self.apply(WebsocketsInfo(**{k: v for k, v in conf.items() if k != "type"}))
                elif conf["type"] in ("webhook", "wh", "http"):
                    self.apply(WebhookInfo(**{k: v for k, v in conf.items() if k != "type"}))
            for conn in self.connections:
                it(Launart).add_component(conn)
        elif key == "cmd_count":
            alconna_config.command_max_count = value
        elif key == "external_dirs":
            log.core.warning("External dirs cannot be changed at runtime, ignored.")
        elif key == "schema":
            self.gen_schema = value

    @property
    def cache(self):
        return it(Launart).get_component(MemcacheService).cache

    def on_message(self, priority: int = 16):
        return le.on(MessageCreatedEvent, priority=priority)

    def ensure_manager(self, manager: Launart):
        self.manager = manager
        manager.add_component(plugin_service)
        manager.add_component(MemcacheService())

        requires(*EntariConfig.instance.prelude_plugin)
        for plug in EntariConfig.instance.prelude_plugin_names:
            load_plugin(plug, prelude=True)
        plugins = EntariConfig.instance.plugin_names
        requires(*plugins)
        for apply, slot in plugin_service._apply.items():
            if slot[1] and apply not in EntariConfig.instance.plugin:
                plugins.append(apply)
        for plug in plugins:
            load_plugin(plug)

        if self.gen_schema and EntariConfig.instance.path.exists():
            EntariConfig.instance.generate_schema(get_plugins())

    async def handle_event(self, account: Account, event: Event):
        try:
            ev = event_parse(account, event)
            if self.ignore_self_message and isinstance(ev, MessageCreatedEvent) and ev.user.id == account.self_id:
                return
            await le.publish(ev)
            return
        except NotImplementedError:
            log.core.warning(f"received unsupported event {event.type}: {event}")

    async def account_hook(self, account: Account, state: LoginStatus):
        await le.publish(AccountUpdate(account, state))

    def run(
        self,
        manager: Launart | None = None,
        *,
        loop: asyncio.AbstractEventLoop | None = None,
        stop_signal: Iterable[signal.Signals] = (signal.SIGINT,),
    ):
        super().run(manager, loop=loop, stop_signal=stop_signal)
        if self._path_scale:
            del sys.path[self._path_scale[0] : self._path_scale[1]]
        if EntariConfig.instance.path.exists():
            EntariConfig.instance.save()
        log.core.info("Entari Shutdown.")

    @classmethod
    def current(cls):
        return it(Launart).get_component(cls)


class EntariProvider(Provider[Entari]):
    priority = 1

    async def __call__(self, context: Contexts):
        return Entari.current()


class LaunartProvider(Provider[Launart]):
    priority = 10

    async def __call__(self, context: Contexts):
        return it(Launart)


class ServiceProviderFactory(ProviderFactory):
    priority = 10

    class _Provider(Provider[Service]):
        def __init__(self, origin: type[Service]):
            super().__init__()
            self.origin = origin

        def validate(self, param: Param):
            anno = get_origin(param.annotation)
            return isinstance(anno, type) and issubclass(anno, self.origin)

        async def __call__(self, context: Contexts):
            try:
                return it(Launart).get_component(self.origin)
            except ValueError:  # Service not loaded yet
                return

    def validate(self, param: Param):
        anno = get_origin(param.annotation)
        if isinstance(anno, type) and issubclass(anno, Service):
            return self._Provider(anno)
        if is_optional(param.annotation, Service):
            args = get_args(param.annotation)
            return self._Provider(args[0])


global_providers.extend([EntariProvider(), LaunartProvider(), ServiceProviderFactory()])  # type: ignore
le.es.set_event_loop(it(asyncio.AbstractEventLoop))



from satori import ChannelType

from ..session import Session


async def direct_message(sess: Session):
    return sess.channel.type == ChannelType.DIRECT


async def public_message(sess: Session):
    return sess.channel.type != ChannelType.DIRECT


async def reply_me(is_reply_me: bool = False):
    return is_reply_me


async def notice_me(is_notice_me: bool = False):
    return is_notice_me


async def to_me(is_reply_me: bool = False, is_notice_me: bool = False):
    return is_reply_me or is_notice_me



from collections.abc import Callable
from pathlib import Path
from tempfile import TemporaryDirectory
from typing_extensions import ParamSpec

from launart import Launart
from launart.service import Phase, Service
from nonestorage import user_cache_dir, user_data_dir

from .config import BasicConfModel, config_model_validate, model_field
from .event.config import ConfigReload
from .plugin import RootlessPlugin, metadata, plugin_config

P = ParamSpec("P")


def _ensure_dir(path: Path) -> None:
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    elif not path.is_dir():
        raise RuntimeError(f"{path} is not a directory")


def _auto_create_dir(func: Callable[P, Path]) -> Callable[P, Path]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Path:
        path = func(*args, **kwargs)
        _ensure_dir(path)
        return path

    return wrapper


class LocalData(Service):
    id = "entari.localdata"

    def __init__(self):
        self.global_path: bool = False
        self.app_name = "entari"
        self.base_dir: str | None = None
        self._temp_dir = TemporaryDirectory(prefix=f"{self.app_name}_")
        super().__init__()

    def _get_base_cache_dir(self) -> Path:
        if self.global_path:
            return user_cache_dir(self.app_name.title()).resolve()
        name = self.base_dir or f".{self.app_name.lstrip('.')}"
        return Path.cwd() / name / "cache"

    def _get_base_data_dir(self) -> Path:
        if self.global_path:
            return user_data_dir(self.app_name.title()).resolve()
        name = self.base_dir or f".{self.app_name.lstrip('.')}"
        return Path.cwd() / name / "data"

    def _get_base_log_dir(self) -> Path:
        if self.global_path:
            return user_data_dir(self.app_name.title()).resolve()
        name = self.base_dir or f".{self.app_name.lstrip('.')}"
        return Path.cwd() / name / "log"

    @_auto_create_dir
    def get_cache_dir(self, name: str | None) -> Path:
        dir_ = self._get_base_cache_dir()
        return (dir_ / name) if name else dir_

    def get_cache_file(self, name: str | None, filename: str) -> Path:
        return self.get_cache_dir(name) / filename

    @_auto_create_dir
    def get_data_dir(self, plugin_name: str | None) -> Path:
        dir_ = self._get_base_data_dir()
        return (dir_ / plugin_name) if plugin_name else dir_

    def get_data_file(self, plugin_name: str | None, filename: str) -> Path:
        return self.get_data_dir(plugin_name) / filename

    @_auto_create_dir
    def get_temp_dir(self) -> Path:
        return Path(self._temp_dir.name)

    def get_temp_file(self, filename: str) -> Path:
        return self.get_temp_dir() / filename

    @property
    def required(self) -> set[str]:
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "blocking", "cleanup"}

    async def launch(self, mgr: Launart):
        async with self.stage("preparing"):
            pass
        async with self.stage("blocking"):
            await mgr.status.wait_for_sigexit()
        async with self.stage("cleanup"):
            self._temp_dir.cleanup()


local_data = LocalData()


class Config(BasicConfModel):
    use_global: bool = model_field(default=False, description="是否使用全局数据目录")
    app_name: str = model_field(default="entari", description="应用名称")
    base_dir: str | None = model_field(default=None, description="基础目录，默认为空，表示使用 `app_name` 作为目录名，")


@RootlessPlugin.apply("localdata", default=True)
def localdata_apply(plg: RootlessPlugin):
    metadata("Data store interface with local directories", ["RF-Tar-Railt <rf_tar_railt@qq.com>"], config=Config)

    plg.service(local_data)
    conf = plugin_config(Config)
    local_data.global_path = conf.use_global
    local_data.app_name = conf.app_name.lower()
    local_data._temp_dir.cleanup()
    local_data._temp_dir = TemporaryDirectory(prefix=f"{local_data.app_name}_")
    local_data.base_dir = conf.base_dir

    @plg.dispatch(ConfigReload)
    def reload_config(event: ConfigReload):
        if event.scope != "plugin":
            return
        if event.key != ".localdata":
            return
        conf1 = config_model_validate(Config, event.value)
        local_data.global_path = conf1.use_global
        local_data.app_name = conf1.app_name.lower()
        local_data._temp_dir.cleanup()
        local_data._temp_dir = TemporaryDirectory(prefix=f"{local_data.app_name}_")
        local_data.base_dir = conf1.base_dir
        return True



from arclet.alconna import Alconna, Args, CommandMeta, Field, Option, namespace
from tarina import lang

from arclet.entari import Session, metadata
from arclet.entari.command import Match, mount

metadata(
    name="语言本地化管理",
    author=[{"name": "RF-Tar-Railt", "email": "rf_tar_railt@qq.com"}],
    description="i18n配置相关功能",
    readme="""
# 语言本地化管理

该插件提供查看和切换语言的功能。

## 使用

- 查看支持的语言列表: `lang list [name?]`, 例如 `lang list entari::plugin/inspect`
- 切换语言: `lang switch [locale?]`，例如 `lang switch zh-CN`
""",
    config=None,
)

with namespace("builtin/lang") as ns:
    ns.disable_builtin_options = {"shortcut", "completion"}

    cmd = mount(
        Alconna(
            "lang",
            Option("list", Args["name?", str], help_text="查看支持的语言列表"),
            Option("switch", Args["locale?", str, Field(completion=lambda: "比如 zh-CN")], help_text="切换语言"),
            meta=CommandMeta("i18n配置相关功能", compact=True),
        ),
    )


@cmd.assign("list")
async def _(name: Match[str], sess: Session):
    try:
        locales = lang.locales_in(name.result) if name.available else lang.locales
    except KeyError:
        await sess.send(f"未能找到 {name.result} 所属的 i18n 目录")
    else:
        await sess.send("支持的语言列表:" + "\n" + "\n".join(f" * {locale}" for locale in locales))


@cmd.assign("switch")
async def _(locale: Match[str], sess: Session):
    if not locale.available:
        resp = await sess.prompt("缺少语言参数，请输入：", timeout=30)
        if resp is None:
            await sess.send("等待语言参数输入超时。")
            return
        _locale = str(resp)
    else:
        _locale = locale.result
    try:
        lang.select(_locale)
    except ValueError as e:
        await sess.send(f"无法切换到语言 {_locale}，请检查语言是否存在。错误信息: {e}")
        return
    else:
        await sess.send(f"切换语言成功: '{_locale}'。")
        return



from __future__ import annotations

from typing import Any

from arclet.alconna import Alconna, command_manager
from arclet.letoderea import define
from arclet.letoderea.provider import TProviders

from ..event.base import MessageCreatedEvent
from ..event.command import CommandExecute
from ..plugin.model import Plugin, PluginDispatcher
from .model import Match, Query
from .provider import AlconnaProviderFactory, AlconnaSuppiler, Assign, MessageJudges, _seminal

exec_pub = define(CommandExecute)
exec_pub.bind(AlconnaProviderFactory())

# fmt: off


class AlconnaPluginDispatcher(PluginDispatcher):
    def __init__(self, plugin: Plugin, command: Alconna, need_reply_me: bool = False, need_notice_me: bool = False, use_config_prefix: bool = True):  # noqa: E501
        plugin._extra.setdefault("commands", []).append((command.prefixes, command.command))
        self.supplier = AlconnaSuppiler(command)
        super().__init__(plugin, MessageCreatedEvent, command.path)
        self.propagators.append(
            MessageJudges(need_reply_me, need_notice_me, use_config_prefix),
        )
        self.propagators.append(self.supplier)
        self.providers.append(AlconnaProviderFactory())

        @plugin.collect
        def dispose():
            command_manager.delete(self.supplier.cmd)
            del self.supplier.cmd
            del self.supplier

    def assign(self, path: str, value: Any = _seminal, or_not: bool = False, priority: int = 16, providers: TProviders | None = None):  # noqa: E501
        assign = Assign(path, value, or_not)
        wrapper = self.register(priority=priority, providers=providers)

        def decorator(func):
            sub = wrapper(func)
            sub.propagate(assign)
            return sub

        return decorator

    def on_execute(self, priority: int = 16, providers: TProviders | None = None):  # noqa: E501
        wrapper = self.plugin._scope.register(priority=priority, providers=providers, publisher=exec_pub)

        def decorator(func):
            sub = wrapper(func)
            sub.propagate(self.supplier)
            return sub

        return decorator

    Match = Match
    Query = Query


def mount(cmd: Alconna, need_reply_me: bool = False, need_notice_me: bool = False, use_config_prefix: bool = True) -> AlconnaPluginDispatcher:  # noqa: E501
    if not (plugin := Plugin.current()):
        raise LookupError("no plugin context found")
    return AlconnaPluginDispatcher(plugin, cmd, need_reply_me, need_notice_me, use_config_prefix)

# fmt: on
